# This file is part of atooms
# Copyright 2010-2024, Daniele Coslovich

"""
The physical system at hand.

The system of interest in a classical atomistic simulations is
composed of interacting point particles, usually enclosed in a
simulation cell. The system may be in contact with a thermostat, a
barostat or a particle reservoir.
"""

import copy
import random
import numpy


def _canonicalize(what):
    """
    Given a descriptor string of a System property, return it
    in canonical form. This is used to canonicalize viewed / dumped
    attributes.
    """
    # Accepts some aliases
    aliases = {
        'box': 'cell.side',
        'pos': 'particle.position',
        'vel': 'particle.velocity',
        'spe': 'particle.species',
        'rad': 'particle.radius'}
    if what in aliases:
        what = aliases[what]

    # If there is no dot, we assume it is a particle property
    if '.' not in what:
        what = 'particle.' + what

    return what

class System(object):

    """System class."""

    def __init__(self, particle=None, cell=None, interaction=None,
                 thermostat=None, barostat=None, reservoir=None,
                 wall=None, molecule=None, N=0, d=3):
        """
        Create a ``System`` with requested objects.

        If provided, `particle` must be a list of `Particle`s.

        If `N` is provided and is an `int`, the system is filled with
        `N` particles starting from a crystalline template in a
        unit-length sided `Cell` in `d` dimensions.

        If `N` is a `dict` specifying a target composition, ex. `{'A':
        N_A, 'B': N_B, ...}`, the system is filled as above but
        setting the species as requested. The assignement is random.
        """
        if particle is None:
            particle = []
        self.particle = particle
        """A list of `Particle` instances."""
        self.interaction = interaction
        self.cell = cell
        self.thermostat = thermostat
        self.barostat = barostat
        self.reservoir = reservoir
        self.wall = wall
        self.molecule = molecule if molecule is not None else []

        # Fill particle list with molecules, if not present.
        # TODO: add switch to avoid or optimize if inefficient
        for molecule in self.molecule:
            for particle in molecule.particle:
                if particle not in self.particle:
                    self.particle.append(particle)

        # Initialize system
        if N != 0:
            from .particle import _lattice
            from .cell import Cell
            assert len(particle) == 0
            if isinstance(N, int):
                N = {'A': N}
            npart = sum(N.values())
            self.particle = _lattice(npart, d)
            self.cell = Cell(numpy.ones(d))
            self.set_composition(N)

        # Internal cache for array dumps
        self._data = None
        self._cache = None

    def __copy__(self):
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__.update(self.__dict__)
        # Note: when making a shallow copy the _data cache
        # may need to be cleared. We might set self._data to None here.
        return result

    def update(self, other, full=False, exclude=None, only=None):
        """
        Update current system attributes in-place using the `other` System
        as source.

        The default behavior is to make deep copies only of all the
        `other` system attributes that are not None, i.e. which are
        set.

        To overwrite all attributes, even those set in `self` but not
        in `other`, use `full=True`.

        The lists `exclude` and `only` can be used to avoid updating
        some system attributes or to update only some system
        attributes, respectively.

        This method can be used by subclasses to deal with performace
        or memory dellocation issues.
        """
        for key in other.__dict__:
            if exclude is not None or only is not None:
                if (exclude is not None and key not in exclude) or \
                   (only is not None and key in only):
                    self.__dict__[key] = copy.deepcopy(other.__dict__[key])
            else:
                if full or other.__dict__[key] is not None:
                    self.__dict__[key] = copy.deepcopy(other.__dict__[key])

        # Internal data dictionary for array dumps
        self._data = None

    @property
    def number_of_dimensions(self):
        """
        Number of spatial dimensions, guessed from the length of
        `particle[0].position`.
        """
        if len(self.particle) > 0:
            return len(self.particle[0].position)
        elif self.cell is not None:
            return len(self.cell.side)
        else:
            return 0

    @property
    def distinct_species(self):
        """Sorted list of distinct chemical species in the system."""
        # TODO: use views to optimize
        from atooms.system.particle import distinct_species as _distinct_species
        return _distinct_species(self.particle)

    @property
    def density(self):
        """
        Density of the system.

        It `cell` is None, an estimate is given.
        """
        if len(self.particle) == 0:
            return 0.0

        if self.cell is None:
            # Estimate density assuming V is the product of the
            # largest distances along each dimension
            volume = 1.0
            for axis in range(len(self.particle[0].position)):
                x_0 = min([p.position[axis] for p in self.particle])
                x_1 = max([p.position[axis] for p in self.particle])
                volume *= (x_1 - x_0)
            return len(self.particle) / volume
        else:
            return len(self.particle) / self.cell.volume

    @density.setter
    def density(self, rho):
        self.set_density(rho)

    def set_density(self, rho):
        """Set the system density to `rho` by rescaling the coordinates."""
        if self.cell is None:
            raise ValueError('cannot compute density without a cell')
        factor = (self.density / rho)**(1./len(self.cell.side))
        for particle in self.particle:
            particle.position *= factor
        self.cell.side *= factor

    @property
    def packing_fraction(self):
        from math import pi
        return pi / 6 * sum([(2 * p.radius)**3 for p in self.particle]) / self.cell.volume

    @property
    def temperature(self):
        """
        Kinetic temperature.
        """
        # TODO: determine translational invariance via some additional attribute.
        if len(self.particle) > 1:
            # Number of degrees of freddom is (N-1)*ndim
            ndof = (len(self.particle)-1) * self.number_of_dimensions
            return 2.0 / ndof * self.kinetic_energy()
        elif len(self.particle) == 1:
            # Pathological case
            return 2.0 / self.number_of_dimensions * self.kinetic_energy()
        else:
            # Empty particle list
            return 0.0

    @temperature.setter
    def temperature(self, T):
        self.set_temperature(T)

    def set_temperature(self, temperature):
        """Reset velocities to a Maxwellian distribution with fixed CM."""
        from .particle import fix_total_momentum
        T = temperature
        for p in self.particle:
            p.maxwellian(T)
        fix_total_momentum(self.particle)
        # After fixing the CM the temperature is not exactly the targeted one
        # Therefore we scale the velocities so as to get to the right T
        T_old = self.temperature
        if T_old > 0.0:
            fac = (T/T_old)**0.5
            for p in self.particle:
                p.velocity *= fac

    @property
    def composition(self):
        """Chemical composition"""
        from .particle import composition
        return dict(composition(self.particle))

    @composition.setter
    def composition(self, value):
        self.set_composition(value)

    def set_composition(self, N):
        """Set the chemical composition as specified in the `N` dictionary."""
        assert sum(N.values()) == len(self.particle), 'composition {} does not match N = {}'.format(N, len(self.particle))
        # Create array of species
        species = []
        for sp in N:
            species += [sp] * N[sp]
        # Randomize the species and assign them
        random.shuffle(species)
        for sp, p in zip(species, self.particle):
            p.species = sp

    @property
    def concentration(self):
        """Chemical concentration"""
        return {k: float(v) / len(self.particle) for k, v in self.composition.items()}

    @concentration.setter
    def concentration(self, value):
        self.set_concentration(value)

    def set_concentration(self, x):
        """Set the chemical concentration as specified in the `x` dictionary."""
        assert abs(sum(x.values()) - 1) < 1e-10, f'concentrations {x} do not sum to 1'
        N = {}
        for species in x:
            N[species] = round(len(self.particle)*x[species])
        if sum(N.values()) != len(self.particle):
            N[species] = sum([N[_] for _ in N if _ != species])
        self.set_composition(N)

    def scale_velocities(self, factor):
        """Scale particles' velocities by `factor`."""
        for p in self.particle:
            p.velocity *= factor

    def kinetic_energy(self, per_particle=False, normed=False):
        """
        Return the total kinetic energy of the system.

        If `per_particle` or `normed` is `True`, return the kinetic
        energy per particle.
        """
        ekin = sum([p.kinetic_energy for p in self.particle])
        if per_particle or normed:
            return ekin / len(self.particle)
        return ekin

    def compute_interaction(self, what=None):
        """
        Compute interaction in the system
        """
        if self.interaction is not None:
            kwargs = {}
            for variable, variable_to_dump in self.interaction.variables.items():
                # Get the optional data type
                dtype = None
                if ':' in variable_to_dump:
                    variable_to_dump, dtype = variable_to_dump.split(':')
                kwargs[variable] = self.dump(variable_to_dump,
                                             view=True, dtype=dtype,
                                             order=self.interaction.order)
            self.interaction.compute(what, **kwargs)

    def potential_energy(self, per_particle=False, normed=False, cache=False):
        """
        Return the total potential energy of the system.

        If `per_particle` or `normed` is `True`, return the potential
        energy per particle.
        """
        if self.interaction is not None:
            if not cache or self.interaction.energy is None:
                # TODO: we should compute 'energy' instead
                self.compute_interaction('forces')
            if per_particle or normed:
                return self.interaction.energy / len(self.particle)
            return self.interaction.energy
        return 0.0

    def total_energy(self, per_particle=False, normed=None, cache=False):
        """
        Return the total energy of the system.

        If `per_particle` or `normed` is `True`, return the total
        energy per particle.
        """
        if normed is not None:
            per_particle = normed
        return self.potential_energy(per_particle, cache=cache) + self.kinetic_energy(per_particle)

    def force_norm(self, per_particle=True, cache=False):
        """
        Return the norm of the force vector.

        If `per_particle` is `True`, return the force norm per particle.
        """
        if self.interaction is not None:
            if not cache or self.interaction.forces is None:
                self.compute_interaction('forces')
            if per_particle:
                return numpy.sum(self.interaction.forces**2)**0.5 / len(self.particle)
            return numpy.sum(self.interaction.forces**2)**0.5
        return 0.0

    def force_norm_square(self, per_particle=True, cache=False):
        """
        Return the squared norm of the force vector.

        If `per_particle` is `True`, return the force squared norm per particle.
        """
        if self.interaction is not None:
            if not cache or self.interaction.forces is None:
                self.compute_interaction('forces')
            if per_particle:
                return numpy.sum(self.interaction.forces**2) / len(self.particle)
            return numpy.sum(self.interaction.forces**2)
        return 0.0

    def virial(self, per_particle=True, cache=False):
        """
        Return the virial of the system.

        If `per_unit_volume` is `True`, return the virial per particle.
        """
        if self.interaction is not None:
            if not cache or self.interaction.virial is None:
                self.compute_interaction('forces')
            if per_particle:
                return self.interaction.virial / len(self.particle)
            return self.interaction.virial
        return 0.0

    @property
    def pressure(self):
        """
        Return the pressure of the system.

        It assumes that `self.interaction` has already been computed.
        """
        if self.thermostat:
            T = self.thermostat.temperature
        else:
            T = self.temperature
        return (len(self.particle) * T + self.interaction.virial / self.number_of_dimensions) / self.cell.volume

    def cm(self, what):
        """General center-of-mass attribute."""
        # It could be implemented in a more general and faster way via dumps
        # but one should pay attention to array ordering and views
        x = numpy.zeros_like(getattr(self.particle[0], what))
        mtot = 0.0
        for p in self.particle:
            x += getattr(p, what) * p.mass
            mtot += p.mass
        return x / mtot

    def center_of_mass(self, what):
        """General center-of-mass attribute."""
        x = self.view(what)
        mass = self.view('mass')
        if x.shape[0] == self.number_of_dimensions:
            x_cm = numpy.dot(x, mass)
        else:
            x_cm = numpy.dot(x.transpose(), mass)
        return x_cm / numpy.sum(mass)

    @property
    def cm_velocity(self):
        """Center-of-mass velocity."""
        return self.cm('velocity')

    @property
    def cm_position(self):
        """Center-of-mass position."""
        return self.cm('position')

    def fix_momentum(self):
        """Subtract out the the center-of-mass motion."""
        vel = self.view('velocity')
        vcm = self.center_of_mass('velocity')
        if vel.shape[0] == self.number_of_dimensions:
            vel[:, :] -= vcm[:, numpy.newaxis]
        else:
            vel[:, :] -= vcm[:, numpy.newaxis].transpose()

    def fix_center_of_mass(self, what='vel'):
        """
        Subtract out the the center-of-mass motion.

        This works with velocity by default, but can be used with any
        attribute, like positions, too.
        """
        var = self.view(what)
        var_cm = self.center_of_mass(what)
        if var.shape[0] == self.number_of_dimensions:
            var[:, :] -= var_cm[:, numpy.newaxis]
        else:
            var[:, :] -= var_cm[:, numpy.newaxis].transpose()

    def fold(self):
        """Fold positions into central cell."""
        for p in self.particle:
            p.fold(self.cell)

    def replicate(self, times, axis):
        """Replicate the system several `times` along `axis`"""
        n = times
        assert n > 1
        npart = len(self.particle)
        L = self.cell.side[axis]
        for p in self.particle:
            p.position[axis] += L/2
        for i in range(1, n):
            for j in range(npart):
                p = copy.deepcopy(self.particle[j])
                p.position[axis] = self.particle[j].position[axis] + i*L
                self.particle.append(p)
        self.cell.side[axis] *= n
        for p in self.particle:
            p.position[axis] -= L * n / 2

    def slab(self, width=1.0, n=0, axis=-1):
        """
        Return slabs orthogonal to `axis` as `System` instances. The slabs'
        thickness is either `width` or such that there are `n` slabs, if
        `n` > 0. The returned systems contain references to the objects of `self`.

        :param width: slab thickness 
        :param axis: axis orthogonal to the slab
        :return: a list of `System` instances
        """
        if n > 0:
            width = self.cell.side[axis] / n
        else:
            n = 1 + int(self.cell.side[axis] / width)
        pos = self.dump('pos')[:, axis]
        idx = (pos[:] + self.cell.side[axis] / 2) / width
        idx = idx.astype('int')
        slabs = [[] for _ in range(n)]
        for i, p in zip(idx, self.particle):
            slabs[i].append(p)
        return [System(particle=slab, cell=self.cell) for slab in slabs]
            
    # Attempt to have a self-contained view method
    # TODO: with attributes that are read only, we should skip the cache
    def __view(self, what, order='C', dtype=None):  # pragma: no cover
        """
        Return a view of system properties specified by `what` as numpy array.

        `what` must be a string of the form `particle.<attribute>` or
        `cell.<attribute>`. Some aliases are allowed like:

        - `pos` (`particle.position`)
        - `vel` (`particle.velocity`)
        - `spe` (`particle.species`)

        If `what` is a System attribute, it is returned.

        The requested particle property is set as
        a view on the corresponding portion of an internal array. This
        allows one to modify the array efficiently and keep the
        particle properties in sync. Numpy arrays must be modified
        in-place to keep everything in sync.

        Particles' coordinates are returned as (N, ndim) arrays if
        `order` is `C` or (ndim, N) arrays if `order` is `F`.

        Examples
        --------

        These numpy arrays are element-wise identical

        .. code-block:: python

           pos = system.dump('particle.position')
           pos = system.dump('position')
           pos = system.dump('pos')

        Collect several properties in a dict

        .. code-block:: python
           data = {what: system.view(what) for what in ['position', 'velocity']}
        """
        # Note: the current design is that view() calls dump() internally
        # and setting the properties in a cache. There is some code
        # duplication, but this way accessing a copy of the data does not
        # interferes with the views. Another approach is that dump() is just view().copy().
        # This will save quite some lines of codes, but one should be careful with order.
        # If the order is different, than we should not change the
        # view but just transpose the copy.

        # It would be better that dump could operate also without an underlying view...
        # For instance, changing a Particle object would break also dump()...
        # But having a cache would be more efficient.

        # TODO: default order should be None. If the cache is not set up, it will be C by default, otherwise we should respect the view.

        assert order in ['C', 'F']
        what = _canonicalize(what)

        # If the attribute to dump is a plain System attribute, return it
        if what in self.__dict__:
            return getattr(self, what)

        # Extract the requested attribute
        assert what.count('.') == 1, f'cannot access nested attribute {what}'
        component, attribute = what.split('.')

        # Setup data cache dictionary for more complex attributes
        if self._cache is None:
            self._cache = {}
            self._cache['data'] = {}
            self._cache['order'] = {}
            self._cache['npart'] = {}

        # Make array of attributes
        assert component in ['particle', 'cell', 'molecule']
        if component == 'particle':
            # We retrieve the view from cache if the array is in the cache
            # and shape and type have not changed.
            # Note: if particles are reassigned without changing N the cache will
            # not be updated. It can be fixed by keeping a list of ids
            # although testing it would bring a overhead.
            if what not in self._cache['data'] or \
                    self._cache['data'][what].dtype != dtype or \
                    self._cache['npart'][what] != len(self.particle) or \
                    self._cache['order'][what] != order:
                # Get the requested data as an array and cache it
                # 1) Using dump internally -> code duplication elsewhere
                # data = self._dump(what, order=order, dtype=dtype)
                # 2) Explicitly here, then dump just make a copy of views
                data = numpy.array([getattr(p, attribute) for p in self.particle], dtype=dtype)
                # We transpose the array if F order is requested
                if order == 'F':
                    data = numpy.transpose(data)
                    data = numpy.asfortranarray(data)

                # TODO: swap order of keys (data <-> what)
                # Set the particle property as a view on the cached array.
                # Note: to check if particle properties and cached array are associated:
                # numpy.may_share_memory(p[0].position, pos[:, 0])
                if order == 'F':
                    for i, p in enumerate(self.particle):
                        setattr(p, attribute, data[..., i])
                if order == 'C':
                    for i, p in enumerate(self.particle):
                        setattr(p, attribute, data[i, ...])

                # Store data in cache
                self._cache['data'][what] = data
                self._cache['order'][what] = order
                self._cache['npart'][what] = len(self.particle)

        if component == 'molecule':
            # TODO: fix this if we ever use this method
            raise ValueError('cannot view molecule attributes, they are read only')
            if what not in self._cache['data']:
                data = numpy.array([getattr(m, attribute) for m in self.molecule])
                self._cache['data'][what] = data

        if component == 'cell':
            if what not in self._cache['data'] or \
               self._cache['data'][what].dtype != dtype:
                # Get the requested data as an array and cache it
                # 1) dump
                # data = self._dump(what, order=order, dtype=dtype)
                # 2) explicitly
                data = numpy.array(getattr(self.cell, attribute), dtype=dtype)
                # Set the particle property as a view on the cached array.
                setattr(self.cell, attribute, data)

                # Store data in cache
                self._cache['data'][what] = data

        return self._cache['data'][what]

    # Attempt to have dump rely on view entirely.
    def __dump(self, what, order='C', dtype=None):
        # TODO: do not modify order/dtype of the view, just cast it if it is different
        # This way the view stays safe and we dump the data efficiently
        return self._view(what, order, dtype).copy()

    # Attempt to have a self-contained dump()
    # def __dump(self, what, order='C', dtype=None):

    #     assert order in ['C', 'F']
    #     what = _canonicalize(what)

    #     # If the attribute to dump is a plain System attribute, return it
    #     if what in self.__dict__:
    #         return copy.deepcopy(getattr(self, what))

    #     # Extract the requested attribute
    #     assert what.count('.') == 1, f'cannot access nested attribute {what}'
    #     component, attribute = what.split('.')

    #     # Make array of attributes
    #     assert component in ['particle', 'cell']
    #     if component == 'particle':
    #         data = numpy.array([getattr(p, attribute) for p in self.particle], dtype=dtype)
    #         # We transpose the array if F order is requested
    #         if order == 'F':
    #             data = numpy.transpose(data)
    #             data = numpy.asfortranarray(data)

    #     if component == 'cell':
    #         data = numpy.array(getattr(self.cell, attribute), dtype=dtype)

    #     return copy.deepcopy(data)

    def dump(self, what=None, order='C', dtype=None, view=False, clear=False, flat=False):
        """
        Return a numpy array with system properties specified by `what`.

        `what` must be a string of the form `particle.<attribute>` or
        `cell.<attribute>`. Some aliases are allowed like:

        - `pos` (`particle.position`)
        - `vel` (`particle.velocity`)
        - `spe` (`particle.species`)

        If `what` is a System attribute, it is returned.

        If `view` is `True`, the requested particle property is set as
        a view on the corresponding portion of the dump array. This
        allows one to modify the dump array efficiently and keep the
        particle properties in sync. Numpy arrays must be modified
        in-place to keep everything in sync.

        If `clear` is True, a new view is created on the requested
        particle property.

        If `flat` is True, the resulting array is flatted following
        `order` using numpy.flatten().

        Particles' coordinates are returned as (N, ndim) arrays if
        `order` is `C` or (ndim, N) arrays if `order` is `F`.

        Examples
        --------

        These numpy arrays are element-wise identical

        .. code-block:: python

           pos = system.dump('particle.position')
           pos = system.dump('position')
           pos = system.dump('pos')
        """
        if isinstance(what, list):
            raise ValueError('list arguments to dump() are not supported, use dict comprehension instead')

        # If the attribute to dump is a plain System attribute, return it
        if what in self.__dict__:
            if view:
                return getattr(self, what)
            else:
                return copy.deepcopy(getattr(self, what))

        # Setup data cache dictionary for more complex attributes
        if self._data is None or clear:
            self._data = {}

        # Return immediately if we only want to clear the dump
        if clear and what is None:
            return

        # Accepts some aliases
        aliases = {
            'box': 'cell.side',
            'pos': 'particle.position',
            'vel': 'particle.velocity',
            'spe': 'particle.species',
            'rad': 'particle.radius'}
        if what in aliases:
            what = aliases[what]

        # We accept particle attributes
        if not what.startswith('particle') and not what.startswith('cell') and \
           not what.startswith('molecule'):
            what = 'particle.' + what
        # Extract the requested attribute
        attr = what.split('.')[-1]

        # Make array of attributes
        if what.startswith('particle'):
            if what in self._data and len(self.particle) == self._data['npart']:
                # TODO: handle change of dtype
                # Get the data in cache if it is present
                # and the number of particles has not changed.
                # Note: if particles are reassigned the cache will
                # not be updated. It can be fixed by keeping a list of ids
                # although testing it would bring a overhead.
                data = self._data[what]
                # If order changes, we transpose the array. This may
                # happen if a trajectory has stored the system in ram
                # with a fortran order and postprocessing asks for a C
                # order array.  This will fail if there are n
                # particles in n dimensions but it is still better
                # than nothing. We must refactor this.
                current_order = 'F' if data.shape[0] == self.number_of_dimensions else 'C'
                if order != current_order:
                    self._data[what] = data.transpose()
                    data = self._data[what]
                # If we are asking a flat copy we must not flatten the cached array
                if not view and flat:
                    data = copy.deepcopy(data).flatten(order=order)
            else:
                # We create a new array
                data = numpy.array([getattr(p, attr) for p in self.particle], dtype=dtype)
                # We transpose the array if F order is requested
                if order == 'F':
                    data = numpy.transpose(data)
                    data = numpy.asfortranarray(data)
                # If view is True, we set the particle property as a
                # view on the dump array. Pay attention of C / F order.
                # To check if particle properties and dump are associated:
                # numpy.may_share_memory(p[0].position, pos[:, 0])
                if view:
                    if not flat:
                        if order == 'C':
                            for i, p in enumerate(self.particle):
                                setattr(p, attr, data[i, ...])
                        if order == 'F':
                            for i, p in enumerate(self.particle):
                                setattr(p, attr, data[..., i])
                    else:
                        data = data.flatten(order=order)
                        shape = getattr(self.particle[0], attr).shape
                        if len(shape) == 2:
                            ndim = self.number_of_dimensions
                            for i, p in enumerate(self.particle):
                                setattr(p, attr, data[i*ndim: (i + 1)*ndim])
                else:
                    # We flatten the array if requested
                    if flat:
                        data = data.flatten(order=order)

        elif what.startswith('molecule'):
            assert not view, 'cannot view molecule attributes, they are read only'
            if what in self._data:
                data = self._data[what]
            else:
                data = numpy.array([getattr(m, attr) for m in self.molecule])

        elif what.startswith('cell'):
            if what in self._data:
                data = self._data[what]
            else:
                data = numpy.array(getattr(self.cell, attr), dtype=dtype)
                if view:
                    setattr(self.cell, attr, data)
        else:
            raise ValueError('Unknown attribute %s' % what)

        if view:
            # Store data in local dict and keep track of the number of particles
            self._data[what] = data
            self._data['npart'] = len(self.particle)
        else:
            data = copy.deepcopy(data)

        return data

    def get(self, what=None, order='C', dtype=None, view=False, clear=False, flat=False):
        """
        Alias of `System.dump()`
        """
        return self.dump(what=what, order=order, dtype=dtype, clear=clear, flat=flat, view=view)

    def view(self, what=None, order='C', dtype=None, clear=False, flat=False):
        """
        Return a view on a system property

        This is an alias of `System.dump(view=True, *args, **kwargs)`
        """
        # TODO: possibly rewrite dump() as view().copy()
        # TODO: if what is None set default internal values of order, flat, dtype (as a dict)
        return self.dump(what=what, order=order, dtype=dtype, clear=clear, flat=flat, view=True)

    def __getitem__(self, key):
        return self.view(key)

    def __setitem__(self, key, value):
        self.view(key)[:] = value

    def __str__(self):
        from .particle import composition
        txt = ''
        if self.particle:
            txt += 'system composed of N={0} particles\n'.format(len(self.particle))
            txt += 'with chemical composition C={}\n'.format(self.composition)
            txt += 'with chemical concentration x={}\n'.format(self.concentration)
        if self.cell:
            txt += 'enclosed in a {0.shape} box at number density rho={1:.6f}\n'.format(self.cell, self.density)
        if self.wall:
            txt += 'surrounded by {} walls\n'.format(len(self.wall))
        if self.thermostat:
            txt += 'in contact with a thermostat at T={0.temperature}\n'.format(self.thermostat)
        if self.barostat:
            txt += 'in contact with a barostat at P={0.pressure}\n'.format(self.barostat)
        if self.reservoir:
            txt += 'in contact with a reservoir at mu={0.chemical_potential}\n'.format(self.reservoir)
        if self.interaction:
            txt += '\n'
            txt += str(self.interaction)
        return txt

    def show(self, backend='matplotlib', **kwargs):
        from .visualize import show_ovito
        from .visualize import show_matplotlib
        from .visualize import show_3dmol
        
        if backend == 'matplotlib':
            _show = show_matplotlib
        elif backend == 'ovito':
            _show = show_ovito
        elif backend == '3dmol':
            _show = show_3dmol
        else:
            raise ValueError(f'unknown {backend} backend for visualization')
        try:
            return _show(self.particle, self.cell, **kwargs)
        except ModuleNotFoundError:
            import warnings
            warnings.warn(f'missing visualization backend {backend}')

    @property
    def species_layout(self):
        """
        Return species layout (A=alphabetical, C=C indexed, F=fortran indexed)
        """
        try:
            species = [int(p.species) for p in self.particle]
        except ValueError:
            # The species cannot be converted to int, thus layout is
            # alphabetical
            # TODO: add periodic table layout 'P'?
            layout = 'A'
        else:
            min_sp = numpy.min(species)
            if min_sp == 0:
                layout = 'C'
            elif min_sp == 1:
                layout = 'F'
            else:
                raise ValueError('Numeric species should start from 0 or 1')
        return layout

    @species_layout.setter
    def species_layout(self, layout):
        """
        Set species layout
        """
        # Sanity checks
        layouts = ['A', 'C', 'F']
        assert layout[0] in layouts, 'layout must be A, C or F (not {})'.format(layout)

        # Do nothing if the layout is already ok
        current_layout = self.species_layout
        if layout == current_layout:
            return

        # Convert to new layout
        import string
        if layout[0] == 'A':
            # We get the index of the species map:
            # - if current layout is F (min_sp=1), we subtract one.
            # - if current layout is C (min_sp=0), we do nothing
            species = [int(p.species) for p in self.particle]
            min_sp = numpy.min(species)
            species_map = string.ascii_uppercase
            for p in self.particle:
                p.species = species_map[int(p.species) - min_sp]
        else:
            # Output layout is numerical (C or F)
            offset = 1 if layout[0] == 'F' else 0
            # Note that distinct_species is sorted alphabetically
            species_list = self.distinct_species
            if current_layout[0] == 'A':
                for p in self.particle:
                    p.species = str(species_list.index(p.species) + offset)
            else:
                # If layout=C, current_layout is F and we subtract 2*offset-1=-1
                # If layout=F, current_layout is C and we add 2*offset-1=+1
                for p in self.particle:
                    p.species = str(int(p.species) + 2*offset - 1)



import os
import bz2
import numpy
from xml.etree import ElementTree
from atooms.trajectory.base import TrajectoryBase, SuperTrajectory
from atooms.system import Particle, Cell, System


class _TrajectoryDynamO(TrajectoryBase):

    """DynamO trajectory format (https://www.dynamomd.com/index.php/tutorial3)"""

    suffix = 'xml'

    def __init__(self, filename, mode='r'):
        super(_TrajectoryDynamO, self).__init__(filename, mode)
        assert mode == 'r', 'cannot write yet'
        self._filename = filename
        if (os.path.splitext(filename)[1][1:].strip() == "bz2"):
            self._file = bz2.BZ2File(filename)
        else:
            self._file = open(filename)
        self.variables = ['particle.species', 'particle.mass',
                          'particle.position', 'particle.velocity']

    def read_timestep(self):
        return 1e-10

    def read_steps(self):
        import re
        base = os.path.basename(self._filename)
        # If the file is named with Snapshot.[0-9]*e.xml* naming
        # scheme then we extract the time from the corresponding
        # output file and convert to a step assuming that frames are
        # stored at regular time intervals. Otherwise the step is 0.
        if base.startswith('Snapshot'):
            # Get the current frame time from output file
            output_file = re.sub('([0-9].)', 'output.\\1', self._filename)
            if (os.path.splitext(output_file)[1][1:].strip() == "bz2"):
                fh = bz2.BZ2File(output_file)
            else:
                fh = open(output_file)
            tree = ElementTree.parse(fh)
            root = tree.getroot()
            time = root.find('Misc').find('Duration').attrib['Time']
            step = round(float(time) / self.timestep)
            fh.close()
        else:
            step = 0
        return [step]

    def read_system(self, frame):
        assert frame == 0
        tree = ElementTree.parse(self._file)
        root = tree.getroot()

        # Cell
        _cell = root.find("Simulation").find('SimulationSize')
        side = numpy.array([_cell.attrib['x'], _cell.attrib['y'], _cell.attrib['z']])

        # Coordinates
        position, velocity = [], []
        for _particle in root.findall('.//Pt'):
            pos = _particle.find('P')
            vel = _particle.find('V')
            position.append([pos.attrib['x'], pos.attrib['y'], pos.attrib['z']])
            velocity.append([vel.attrib['x'], vel.attrib['y'], vel.attrib['z']])

        # Species
        _species = root.find("Simulation").find("Genus").findall('Species')
        if len(_species) == 1:
            species, mass = len(position) * ['A'], len(position) * [1.0]
        else:
            species, mass = [], []
            for sp in _species:
                assert sp.find('IDRange').attrib['Type'] == 'Ranged'
                begin = int(sp.find('IDRange').attrib['Start'])
                end = int(sp.find('IDRange').attrib['End'])
                for _ in range(begin, end+1):
                    species.append(sp.attrib['Name'])
                    mass.append(sp.attrib['Mass'])

        # Pack everything into a System
        cell = Cell(side)
        particle = [Particle(species=s, mass=float(m),
                             position=numpy.array(p, dtype=float),
                             velocity=numpy.array(v, dtype=float))
                    for p, m, s, v in zip(position, mass, species, velocity)]
        return System(particle, cell)

    def close(self):
        self._file.close()


class TrajectoryDynamO(TrajectoryBase):

    """DynamO trajectory format (https://www.dynamomd.com/index.php/tutorial3)"""

    def __init__(self, path, mode='r'):
        import os
        import glob

        super(TrajectoryDynamO, self).__init__(path, mode)
        if os.path.isdir(path):
            files = glob.glob(path + '/Snapshot.[0-9]*.*')
            self._th = SuperTrajectory(files, _TrajectoryDynamO)
        else:
            self._th = _TrajectoryDynamO(path)
        self.steps = self._th.steps

    def read_timestep(self):
        return self._th.read_timestep()

    def read_system(self, frame):
        return self._th.read_system(frame)



#!/usr/bin/env python

import unittest
from atooms import core
from atooms.trajectory import utils
from atooms.trajectory import TrajectoryXYZ


class Test(unittest.TestCase):

    def setUp(self):
        pass

    def test_block_size_1(self):
        more = [0, 1, 2, 3, 4]
        self.assertEqual(utils.get_block_size(more), 1)
        self.assertEqual(utils.check_block_size(more, utils.get_block_size(more)), None)

    def test_block_size_2(self):
        more = [0, 1, 10, 11, 20, 21]
        self.assertEqual(utils.get_block_size(more), 2)
        self.assertEqual(utils.check_block_size(more, utils.get_block_size(more)), more)

    def test_block_size_3(self):
        more = [0, 1, 10, 12, 20, 30]
        self.assertEqual(utils.get_block_size(more), len(more))

    def test_block_size_3(self):
        more = [0, 1, 2, 4, 8, 16]
        more += [32, 33, 34, 36, 40, 48]
        self.assertEqual(utils.get_block_size(more), 6)
        self.assertEqual(utils.check_block_size(more, utils.get_block_size(more)), more)

    def test_block_size_4(self):
        more = [0, 1, 2, 4, 8, 16]
        more += [32 + i for i in more]
        self.assertEqual(utils.get_block_size(more), 6)
        self.assertEqual(utils.check_block_size(more, utils.get_block_size(more)), more)

    def test_block_size_5(self):
        more = [0, 1, 2, 4, 8, 16, 24, 32]
        more += [40 + i for i in more]
        self.assertEqual(utils.get_block_size(more), 8)
        self.assertEqual(utils.check_block_size(more, utils.get_block_size(more)), more)

    def test_cell_variable(self):
        finp = '/tmp/test_utils.xyz'
        with open(finp, 'w') as fh:
            fh.write("""\
1
step:0 cell:1.0,1.0,1.0
A 1.0 -1.0 0.0
1
step:1 cell:2.0,1.0,1.0
A 1.0 -1.0 0.0
1
step:2 cell:1.0,1.0,1.0
B 1.0 -1.0 0.0
1
step:3 cell:1.0,1.0,1.0
A 1.0 -1.0 0.0
""")
        with TrajectoryXYZ(finp) as th:
            self.assertTrue(utils.is_cell_variable(th, tests=-1))
            self.assertFalse(utils.is_cell_variable(th, tests=1))
            self.assertTrue(utils.is_cell_variable(th, tests=2))

            # Test utility function
            from atooms.trajectory.utils import dump
            pos = dump(th, what='pos')
            from atooms.trajectory.utils import is_semigrandcanonical, is_grandcanonical
            self.assertFalse(is_grandcanonical(th))
            self.assertTrue(is_semigrandcanonical(th, tests=3))

    def test_formats(self):
        from atooms.trajectory.utils import formats
        self.assertTrue('xyz' in formats())

    def test_file_index(self):
        from atooms.trajectory import utils
        finp = '/tmp/test_utils.xyz'
        with open(finp, 'w') as fh:
            fh.write("""\
2
step:0 cell:1.0,1.0,1.0
A 1.0 -1.0 0.0
A 1.0 -1.0 0.0
3
step:1 cell:2.0,1.0,1.0
A 1.0 -1.0 0.0
A 1.0 -1.0 0.0
A 1.0 -1.0 0.0
3
step:2 cell:1.0,1.0,1.0
A 1.0 -1.0 0.0
A 1.0 -1.0 0.0
A 1.0 -1.0 0.0
""")
        with open(finp) as fh:
            utils.file_index(fh)

    def test_dumps(self):
        from atooms.core import utils
        utils.report_command('cmd', {'a': 1, 'b': None, 'c': True}, '', None)
        utils.report_parameters({'x': 1.0}, None, '1.0.0')

    def test_timings(self):
        import sys
        import os
        def f(x): return x
        f = core.utils.clockit(f, output=open(os.devnull, 'w'))
        f(1)
        t = core.utils.Timer()
        t.start()
        t.stop()
        t.cpu_time
        t.wall_time
        with core.utils.Timer(output=open(os.devnull, 'w')) as _:
            pass

    def test_slice(self):
        self.assertTrue(core.utils.fractional_slice(0.2, 0.5, None, 10) == slice(2, 5, None))
        self.assertTrue(core.utils.fractional_slice(0, 0.5, None, 10) == slice(0, 5, None))
        self.assertTrue(core.utils.fractional_slice(0.5, 10, None, 10) == slice(5, 10, None))
        self.assertTrue(core.utils.fractional_slice(0.2, 10, None, 10) == slice(2, 10, None))

    def test_parser(self):
        import argparse
        parser = argparse.ArgumentParser()
        parser = core.utils.add_first_last_skip(parser)

    def test_ordered_set(self):
        from atooms.system.particle import Particle
        particle = [Particle(species='A'), Particle(species='C')]
        periodic_table = core.utils.OrderedSet()
        periodic_table.update([p.species for p in particle])
        particle = [Particle(species='A'), Particle(species='D')]
        periodic_table.update([p.species for p in particle])
        self.assertEqual(periodic_table.index('C'), 1)
        self.assertEqual(periodic_table[0], 'A')
        self.assertEqual(list(periodic_table), ['A', 'C', 'D'])

    def tearDown(self):
        from atooms.core.utils import rmf
        rmf('/tmp/test_utils*')


if __name__ == '__main__':
    unittest.main()



# This file is part of atooms
# Copyright 2010-2024, Daniele Coslovich

"""Simulation cell."""

import numpy
import warnings
from atooms.core import ndim as _ndim


class Cell(object):

    def __init__(self, side=None, center=None, periodic=None):
        if side is None:
            self.side = numpy.zeros(_ndim)
        else:
            self.side = numpy.asarray(side, dtype=numpy.float64)
        # By default, the center is at the origin of the reference frame
        if center is None:
            self.center = numpy.zeros_like(self.side)
        else:
            self.center = numpy.asarray(center, dtype=numpy.float64)
        self.shape = 'cubic'

        # Periodic boundary conditions apply separately on each axis
        if periodic is None:
            self.periodic = numpy.ndarray(self.side.size, dtype=bool)
            self.periodic[:] = True
        else:
            self.periodic = numpy.asarray(periodic, dtype=bool)

    @property
    def volume(self):
        """Volume of the cell"""
        return numpy.prod(self.side)

    @property
    def origin(self):
        warnings.warn('cell origin is deprecated', DeprecationWarning)
        return numpy.zeros(_ndim)



import logging
import numpy

from .xyz import TrajectoryXYZ
from .utils import gopen, _PossiblyZippedFile
from atooms.core.utils import tipify
from atooms.system.particle import Particle
from atooms.system.cell import Cell
from atooms.system import System

log = logging.getLogger(__name__)


class TrajectoryEXYZ(TrajectoryXYZ):

    """
    Extended XYZ layout (https://github.com/libAtoms/extxyz)
    """

    suffix = 'xyz'
    callback_read = {}

    def __init__(self, filename, mode='r'):
        super(TrajectoryEXYZ, self).__init__(filename, mode)

        # Trajectory file handle
        self.precision = 12
        self.alias = {'pos': 'position',
                      'vel': 'velocity'}
        self.variables = ['species', 'pos']
        # Internal index of lines via seek and tell.
        if self.mode == 'r':
            self._setup_index()
            assert len(self._index_frame) > 0, 'empty file {}'.format(self._file)
            assert len(self._index_header) > 0, 'empty file {}'.format(self._file)
            # Read metadata
            self.metadata = self._read_comment(0)
            # Note: currently, changing variables in read mode will
            # not affect which attributes are parsed
            self.variables = [key for key, _, _ in self.metadata['Properties']]

    def _read_comment(self, frame):
        """
        Internal xyz method to get header metadata from comment line of
        given `frame`.
        """
        # Go to line and skip Npart info
        self._file.seek(self._index_header[frame])
        npart = int(self._file.readline())
        data = self._file.readline()
        meta = {}

        # We first gather all keys
        keys = []
        for entry in data.split("=")[:-1]:
            keys.append(entry.split()[-1])

        # Now we extract the values for each key using a dynamic regexp
        import re
        regexp = ''
        for key in keys:
            regexp += '{}=(.*)'.format(key)
        match = re.match(regexp, data)
        for i, key in enumerate(keys):
            meta[key] = match.groups()[i].strip()

        # The Properties key is special so we deal with it first
        entries = meta['Properties'].split(':')
        assert len(entries) % 3 == 0, len(entries)
        properties = []
        for i in range(0, len(entries), 3):
            key, fmt, ndims = entries[i: i+3]
            properties.append((key, fmt, ndims))

        meta['Properties'] = properties

        # Go through keys, listify and tipify them
        for key in meta:
            if key != 'Properties':
                if meta[key].startswith('"'):
                    meta[key] = meta[key].strip('"')
                    meta[key] = [tipify(_) for _ in meta[key].split()]
                else:
                    meta[key] = tipify(meta[key])

        return meta

    def read_steps(self):
        """Find steps list."""
        steps = []
        for frame in range(len(self._index_frame)):
            meta = self._read_comment(frame)
            try:
                steps.append(meta['Step'])
            except KeyError:
                # If no step info is found, we add steps sequentially
                steps.append(frame+1)
        return steps

    def read_system(self, frame):
        # Read metadata of this frame
        meta = self._read_comment(frame)

        # Get number of particles
        self._file.seek(self._index_header[frame])
        npart = int(self._file.readline())

        # Read frame now
        self._file.seek(self._index_frame[frame])
        particle = []
        for _ in range(npart):
            p = Particle()
            data = self._file.readline().split()
            i = 0
            for key, fmt, ndims in meta['Properties']:
                ndims = int(ndims)
                if key in self.alias:
                    key = self.alias[key]
                if ndims == 1:
                    if fmt == 'R':
                        setattr(p, key, float(data[i]))
                    elif fmt == 'I':
                        setattr(p, key, int(data[i]))
                    elif fmt == 'S':
                        setattr(p, key, data[i])
                    else:
                        raise ValueError('unknown format key')
                else:
                    if fmt == 'R':
                        setattr(p, key, numpy.array(data[i: i+ndims], dtype=numpy.float64))
                    elif fmt == 'I':
                        setattr(p, key, numpy.array(data[i: i+ndims], dtype=numpy.int64))
                    elif fmt == 'S':
                        setattr(p, key, numpy.array(data[i: i+ndims]))
                    else:
                        raise ValueError('unknown format key')
                i += ndims
            particle.append(p)

        side = meta["Lattice"]
        # TODO: remove hard coded
        cell = Cell([side[0], side[4], side[8]])
        return System(particle, cell)

    def read_timestep(self):
        for key in ['dt', 'Dt', 'timestep', 'Timestep']:
            if key in self.metadata:
                return self.metadata[key]
        return 1.0

    def _comment(self, step, system):
        # TODO: improve
        L = []
        for i in range(len(system.cell.side)):
            for j in range(len(system.cell.side)):
                if i == j:
                    L.append(system.cell.side[i])
                else:
                    L.append(0.0)

        # Reformat properties
        # TODO: improve
        _properties = ''
        for i in range(len(self._properties)):
            _properties += ':'.join([str(_) for _ in self._properties[i]]) + ':'
        _properties = _properties.strip(':')

        line = 'Properties={} '.format(_properties)
        line += 'Lattice="{}" '.format(' '.join([str(_) for _ in L]))
        line += 'Timestep={} '.format(self.timestep)
        line += 'Step={} '.format(step)
        return line.strip()

    def write_system(self, system, step):
        from atooms.core.utils import is_array

        def detect_format(arg):
            if isinstance(arg, float):
                return 'R'
            elif isinstance(arg, int):
                return 'I'
            else:
                return 'S'

        properties = []
        for variable in self.variables:
            attr = variable.split('.')[1]
            # Guess attribute features (ndim, fmt)
            val = getattr(system.particle[0], attr)
            if is_array(val):
                ndim, fmt = len(val), detect_format(val[0])
            else:
                ndim, fmt = 1, detect_format(val)
            properties.append([attr, fmt, ndim])
        # TODO: this should be aliased to stick with shortcuts
        self._properties = properties

        # Write header
        self._file.write('{}\n'.format(len(system.particle)))
        self._file.write(self._comment(step, system) + '\n')

        # Formatters
        alias_fmt = {'S': '{} ',
                     'R': '{{:.{precision}f}} '.format(precision=self.precision),
                     'I': '{} '}

        # Write particles
        for p in system.particle:
            line = ''
            for i in range(len(properties)):
                key, fmt, ndims = properties[i]
                val = getattr(p, key)
                if ndims == 1:
                    line += alias_fmt[fmt].format(val)
                else:
                    # Note: numpy.array2string is MUCH slower
                    line += ' '.join([alias_fmt[fmt].format(x) for x in val])
            self._file.write(line.strip() + '\n')



# This file is part of atooms
# Copyright 2010-2024, Daniele Coslovich

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

"""
Global variables.
"""

__author__ = "Daniele Coslovich <daniele.coslovich@umontpellier.fr>"

from ._version import __version__ as __bare_version

try:
    from ._commit import __commit__, __date__
    __version__ = '%s+%s (%s)' % (__bare_version, __commit__, __date__)
except ImportError:
    __commit__ = ""
    __date__ = ""
    __version__ = __bare_version

ndim = 3
"""Number of spatial dimensions."""



import numpy
from .base import TrajectoryBase
from atooms.system import System, Particle


class TrajectoryCSV(TrajectoryBase):

    def __init__(self, filename, mode='r'):
        super(TrajectoryCSV, self).__init__(filename, mode)
        self._file = open(self.filename, self.mode)
        self.precision = 4
        self._ndim = 3
        self._sep = None
        self._order = 'C'
        if mode == 'r':
            self._setup()

    def _setup(self):
        self._file_index = []
        self._file.seek(0)
        while True:
            line = self._file.tell()
            data = self._file.readline()
            if not data:
                break
            else:
                self._file_index.append(line)
        self._file.seek(0)
        self.steps = range(len(self._file_index))

    def read_system(self, frame):
        self._file.seek(self._file_index[frame])
        data = self._file.readline().strip()
        pos = numpy.array([float(_) for _ in data.split(self._sep)])
        npart = len(pos) // self._ndim
        pos = pos.reshape((npart, self._ndim))
        system = System()
        for i in range(npart):
            # p = Particle(position=pos[i*self._ndim: (i+1)*self._ndim])
            p = Particle(position=pos[i, :])
            p.position_unfolded = p.position
            system.particle.append(p)
        return system

    def write_system(self, system, step):
        sep = ' '
        fmt = '{:.' + str(self.precision) + 'g}'
        fmt_func = lambda x: sep.join([fmt.format(xi) for xi in x])
        pos = system.dump('position', flat=True, order=self._order)
        if numpy.version.version[0] == '1':
            numpy.set_string_function(fmt_func, False)
            self._file.write(str(pos) + '\n')
            numpy.set_string_function(None, False)
        else:
            numpy.set_printoptions(override_repr=fmt_func)
            self._file.write(repr(pos) + '\n')
            numpy.set_printoptions()

    def close(self):
        self._file.close()



# This file is part of atooms
# Copyright 2010-2024, Daniele Coslovich

"""
Abstract factory class for trajectories.

The callable `TrajectoryFactory` returns the appropriate trajectory
class based on either explicit rules or suffix euristics. The list of
available trajectory classes can be updated at run-time using the
`update` method.

Examples
--------

- Reading an hdf5-format trajectory via the factory

.. code-block: python

   Trajectory = TrajectoryFactory()
   t = Trajectory('input.h5')

is equivalent to

.. code-block: python

   t = TrajectoryHDF5('input.h5')

- Ask for a specific trajectory format

.. code-block: python

   Trajectory = TrajectoryFactory()
   t = Trajectory('input.xyz', fmt='rumd')

This is equivalent to

.. code-block: python

   t = TrajectoryRUMD('input.xyz')
"""

import os
import sys
import re
import inspect


# Note: trajectories should implement a method to check if a file is
# of their own format or not, to avoid relying on suffix check out
# http://stackoverflow.com/questions/456672/class-factory-in-python.
# However, one would have to deal with conflicts anyway, so the
# settings in update() are fine imo.

class TrajectoryFactory(object):

    def __init__(self):
        self.formats = {}
        self.suffixes = {}
        self.callbacks = []

    def _add(self, trj_class, trj_name, overwrite=True):
        """Atomically add one trajectory class to the factory"""
        # We extract the name of the trajectory and lowercase it
        # We expect names of the format [Tag1]Trajectory<Tag2>, where
        # <Tag2> is obligatory and Tag1 is optional. The trajectory key
        # is then Tag1+Tag2 lowercased.
        s = re.search(r'([a-zA-Z0-9]*)Trajectory([a-zA-Z0-9]*)', trj_name)
        if len(s.group(2)) == 0:
            return
        fmt = (s.group(1) + s.group(2)).lower()
        # Always add the new class to the dict of trajectory formats
        # but only add its suffix if not already present (unless we overwrite)
        self.formats[fmt] = trj_class
        if trj_class.suffix not in self.suffixes or overwrite:
            self.suffixes[trj_class.suffix] = trj_class

    def add(self, trajectory_class):
        """Add a trajectory class to the factory"""
        trajectory_name = str(trajectory_class)
        self._add(trajectory_class, trajectory_name)

    def update(self, module, overwrite=True):
        """Update factory parsing all trajectory classes from module"""
        classes = inspect.getmembers(sys.modules[module], inspect.isclass)
        trajectories = [c for c in classes if 'Trajectory' in c[0]]
        for trj_name, trj_class in trajectories:
            if trj_name == 'TrajectoryFactory':
                continue
            self._add(trj_class, trj_name, overwrite)

    def register_callback(self, cbk, *args, **kwargs):
        """
        Register a callback `cbk` to be applied when reading a frame.

        The callback is registered in the Trajectory instance when
        calling the factory, see __call__().
        """
        if (cbk, args, kwargs) not in self.callbacks:
            self.callbacks.append((cbk, args, kwargs))

    def __call__(self, filename, mode='r', fmt=None):
        if fmt is not None:
            return self.formats[fmt](filename, mode)

        th = None
        suffix = os.path.splitext(filename)[-1].replace('.', '')
        if suffix in self.suffixes:
            th = self.suffixes[suffix](filename, mode)
        else:
            # Fallback to hdf5
            if 'hdf5' in self.formats:
                try:
                    th = self.formats['hdf5'](filename, mode)
                except OSError:
                    pass
        if th is not None:
            # Register the callbacks in the actual trajectory instance
            for cbk, args, kwargs in self.callbacks:
                th.register_callback(cbk, *args, **kwargs)
            return th
        else:
            raise ValueError('unknown file format for %s' % filename)



# This file is part of atooms
# Copyright 2010-2024, Daniele Coslovich

"""
Systems are composed by particles enclosed in a simulation cell,
possibly in contact with a reservoir.
"""

from .system import System
from .particle import Particle
from .molecule import Molecule
from .cell import Cell
from .reservoir import Thermostat, Barostat, Reservoir
from .interaction import Interaction, InteractionBase, \
    InteractionWall, InteractionField
from .wall import Wall



# This file is part of atooms
# Copyright 2010-2024, Daniele Coslovich

"""
Simulation framework for systems of interacting particles.

`atooms` provides a generic simulation interface that abstracts out
most of the common parts of particle-based simulations. It uses
callbacks to analyze and process simulation data on the fly.
"""

from .core import Simulation
from .observers import *
from .reports import *



#!/usr/bin/env python

import os
import unittest
import numpy

from atooms.core.utils import rmd, rmf
from atooms.system import System, Particle, Cell
from atooms.trajectory import Unfolded
from atooms.trajectory import TrajectoryXYZ, TrajectorySimpleXYZ, TrajectoryRUMD, Trajectory
from atooms.trajectory.base import TrajectoryBase
import atooms.trajectory as trj


def almost_equal(x, y, rtol):
    try:
        numpy.testing.assert_allclose(x, y, rtol=rtol)
        return True
    except AssertionError as exc:
        return False

def _equal(system1, system2, ignore=None, verbose=True, precision=1e-10):
    check = {}
    check['npart'] = len(system1.particle) == len(system2.particle)
    check['side'] = all(system1.cell.side == system1.cell.side)
    for p1, p2 in zip(system1.particle, system2.particle):
        # There is roundoffs with gsd! It is surprising because it is a binary format. It should be up to machine precision... difference is 10 places
        # check['position'] = all(p1.position == p2.position)
        check['position'] = almost_equal(p1.position, p2.position, rtol=precision)
        check['velocity'] = almost_equal(p1.velocity, p2.velocity, rtol=precision)
        check['mass'] = p1.mass == p2.mass
        check['species'] = p1.species == p2.species
    for key in check:
        if ignore is not None and key in ignore:
            continue
        if not check[key]:
            if verbose:
                print(key, 'differs')
            return False
    return True

def _difference(system1, system2, ignore=None, precision=1e-10):
    check = {}
    check['npart'] = len(system1.particle) == len(system2.particle)
    check['side'] = all(system1.cell.side == system1.cell.side)
    for p1, p2 in zip(system1.particle, system2.particle):
        check['position'] = almost_equal(p1.position, p2.position, rtol=precision)
        check['velocity'] = almost_equal(p1.velocity, p2.velocity, rtol=precision)
        check['mass'] = p1.mass == p2.mass
        check['species'] = p1.species == p2.species
    diffs = []
    for key in check:
        if ignore is not None and key in ignore:
            continue
        if not check[key]:
            diffs.append(key)
    return diffs


def _rename_species(particle, db):
    for p in particle:
        p.species = db[p.species]
    return particle


inpfile = '/tmp/test_trajectory'
inpdir = '/tmp/test_trajectory.d'

class Test(unittest.TestCase):

    def setUp(self):
        import copy
        particle = [Particle(position=[0.1, 0.2, 0.3], velocity=[0.3, 0.2, 0.1], species='A', mass=1.0),
                    Particle(position=[1.3, 1.2, 1.1], velocity=[1.1, 1.2, 1.3], species='B', mass=2.0),
                    ]
        cell = Cell([2.0, 2.0, 2.0])
        self.system = []
        self.system.append(System(copy.deepcopy(particle), cell))
        self.system.append(System(copy.deepcopy(particle), cell))
        self.inpfile = '/tmp/test_trajectory'
        self.inpdir = '/tmp/test_trajectory.d'
        from atooms.core.utils import mkdir
        mkdir(self.inpdir)

    def _read_write(self, cls, path=inpfile, ignore=None, precision=1e-10):
        """Read and write"""
        with cls(path, 'w') as th:
            th.write_timestep(1.0)
            for i, system in enumerate(self.system):
                th.write(system, i)
        with cls(path, 'r') as th:
            self.assertEqual(th.timestep, 1.0)
            for i, system in enumerate(th):
                self.assertTrue(_equal(self.system[i], system, ignore, precision=precision))
                self.assertTrue(self.system[i].__class__ is system.__class__)

    def _read_write_ram(self, cls, ignore=None, precision=1e-10):
        """Read and write"""
        with cls() as th:
            th.write_timestep(1.0)
            for i, system in enumerate(self.system):
                th.write(system, i)
            self.assertEqual(th.timestep, 1.0)
            for i, system in enumerate(th):
                self.assertTrue(_equal(self.system[i], system, ignore, precision=precision))
                self.assertTrue(self.system[i].__class__ is system.__class__)

    def _read_write_fields(self, cls, write_fields=None, read_fields=None, path=inpfile, ignore=None, fail=None, precision=1e-10):
        """Read and write with fields"""
        # Write
        # try:
        #     th = cls(path, 'w', fields=write_fields)
        # except TypeError:
        th = cls(path, 'w')
        if write_fields is not None:
            th.variables = write_fields
        th.write_timestep(1.0)
        for i, system in enumerate(self.system):
            th.write(system, i)
        th.close()

        # Read
        # try:
        #     th = cls(path, fields=read_fields)
        # except TypeError:
        th = cls(path)
        if read_fields is not None:
            th.variables = read_fields
        self.assertEqual(th.timestep, 1.0)
        for i, system in enumerate(th):
            if fail is not None:
                self.assertTrue(set(_difference(self.system[i], system, ignore, precision=precision)) == set(fail))
            else:
                self.assertTrue(_equal(self.system[i], system, ignore, verbose=False, precision=precision))
            self.assertTrue(self.system[i].__class__ is system.__class__)
        th.close()

    def _write(self, cls, path=None):
        """Write only"""
        if path is None:
            path = self.inpfile
        with cls(path, 'w') as th:
            th.write_timestep(1.0)
            for i, system in enumerate(self.system):
                th.write(self.system[i], i)

    def _slice(self, cls, path=None):
        """Write only"""
        if path is None:
            path = self.inpfile
        with cls(path, 'w') as th:
            for i, system in enumerate(self.system):
                th.write(system, i)
            from atooms.trajectory.decorators import Sliced
            ts = Sliced(th, slice(None, None, 2))
            self.assertEqual(len(th), 2)
            self.assertEqual(len(ts), 1)
            # This will fail, we cannot slice twice yet
            # ts = Sliced(ts, slice(None, None, 1))
            # print len(ts)

    def _append(self, cls, path=inpfile, ignore=None):
        """Read and write"""
        with cls(path, 'w') as th:
            for system in self.system:
                th.append(system)
        with cls(path) as th:
            for i, system in enumerate(th):
                self.assertTrue(_equal(self.system[i], system, ignore))
                self.assertTrue(self.system[i].__class__ is system.__class__)

    def _convert(self, cls_inp, cls_out, path=None, ignore=None):
        """Write then convert"""
        if path is None:
            path = self.inpfile
        fout = self.inpfile + '.out'
        with cls_inp(path, 'w') as th:
            th.write_timestep(1.0)
            for i, system in enumerate(self.system):
                th.write(self.system[i], i)

        with cls_inp(path) as th:
            with th.copy(cls_out, fout) as _:
                pass

        if isinstance(cls_out, str):
            th = trj.Trajectory(fout, fmt=cls_out)
        else:
            th = cls_out(fout)
        self.assertEqual(th.timestep, 1.0)
        self.assertTrue(len(th.steps), len(self.system))
        for i, system in enumerate(th):
            self.assertTrue(_equal(self.system[i], system, ignore))
        th.close()

    def test_xyz(self):
        # TODO: mass is not written by xyz
        # Velocity is not written in default xyz format, so we ignore it
        self._read_write(trj.TrajectoryXYZ, ignore=['mass', 'velocity'])
        self._convert(trj.TrajectoryXYZ, trj.TrajectoryXYZ, ignore=['mass', 'velocity'])
        self._convert(trj.TrajectoryXYZ, 'xyz', ignore=['mass', 'velocity'])
        self._append(trj.TrajectoryXYZ, ignore=['mass', 'velocity'])
        self._slice(trj.TrajectoryXYZ)
        # Check that when requesting to write the velocity, we actually write it and read it back automatically
        self._read_write_fields(trj.TrajectoryXYZ, write_fields=['species', 'position', 'velocity'], ignore=['mass'])
        # This must fail: writing positions but not reading them
        self._read_write_fields(trj.TrajectoryXYZ, write_fields=['species', 'position', 'velocity'],
                                read_fields=['species', 'velocity'], ignore=['mass'], fail=['position'])

    def test_simple_xyz(self):
        # Mass and velocity is not written in default xyz format, so we ignore it
        self._read_write(trj.TrajectorySimpleXYZ, ignore=['mass', 'velocity'])
        self._append(trj.TrajectorySimpleXYZ, ignore=['mass', 'velocity'])

    def test_ram(self):
        self._read_write_ram(trj.TrajectoryRam)
        self._append(trj.TrajectoryRam, path=None)

    def test_ram_full(self):
        self._read_write_ram(trj.ram.TrajectoryRamFull)
        self._append(trj.ram.TrajectoryRamFull, path=None)

    def test_ram_from_data(self):
        # Get positions and species as numpy array
        pos, spe, box = [], [], []
        with trj.TrajectoryXYZ('data/ka_N150_rho1.200.xyz') as th:
            for s in th:
                pos.append(s.dump('position'))
                spe.append(s.dump('particle.species'))
                box.append(s.dump('box'))
            steps = th.steps
        # Store trajectory in memory from data
        data = {'pos': pos, 'spe': spe, 'box': box}
        th = trj.TrajectoryRam(data)
        th.steps = steps
        for i in range(len(th)):
            self.assertTrue(numpy.all(th[i].dump('pos') == pos[i]))
            self.assertTrue(numpy.all(th[i].dump('spe') == spe[i]))
            self.assertTrue(numpy.all(th[i].dump('box') == box[i]))

    def test_ram_light(self):
        self._read_write_ram(trj.ram.TrajectoryRamLight)
        self._append(trj.ram.TrajectoryRamLight, path=None)

    def test_ram_light_custom(self):
        # Test custom attributes
        with trj.ram.TrajectoryRamLight() as th:
            for i, system in enumerate(self.system):
                for p in system.particle:
                    p.custom_field = numpy.ones(3)
                th.write(system, i)
            for system in th:
                x = system.view('particle.custom_field')
                self.assertEqual(x.shape, (2, 3))

    def test_hdf5(self):
        try:
            import h5py
        except ImportError:
            self.skipTest('missing hdf5')

        self._read_write(trj.TrajectoryHDF5)
        self._read_write_fields(trj.TrajectoryHDF5, write_fields=['species', 'position', 'velocity'], read_fields=['species', 'position', 'velocity'])
        # This must fail: writing velocities but not reading them
        self._read_write_fields(trj.TrajectoryHDF5, write_fields=['species', 'position'], read_fields=['species', 'position', 'velocity'], fail=['velocity'])
        # Velocity is not kept in conversion to xyz
        self._convert(trj.TrajectoryXYZ, 'hdf5', ignore=['mass', 'velocity'])

    def test_gsd(self):
        try:
            import gsd
            from atooms.trajectory.gsd import TrajectoryGSD
        except ImportError:
            self.skipTest('missing gsd')

        self._read_write(trj.TrajectoryGSD, ignore=['mass', 'velocity'], precision=1e-7)
        # self._convert(trj.TrajectoryGSD, 'gsd', ignore=['mass', 'velocity'])
        # self._read_write_fields(trj.TrajectoryGSD, write_fields=['species', 'position', 'velocity'], read_fields=['species', 'position', 'velocity'], ignore=['mass'])
        # This must fail: writing velocities but not reading them
        # self._read_write_fields(trj.TrajectoryGSD, write_fields=['species'], read_fields=['species', 'position'], ignore=['mass'], fail=['velocity'])

    def test_rumd(self):
        # RUMD uses integer ids for chemical species. They should be integers.
        for s in self.system:
            s.particle = _rename_species(s.particle, {'A': '0', 'B': '1'})
        self._read_write(trj.TrajectoryRUMD)
        # TODO: all these fail without changing _optimize_fields() in setup_fields() so as to return a new list
        # self._read_write_fields(trj.TrajectoryRUMD, read_fields=['species', 'position', 'velocity'], ignore=['mass'])
        # self._read_write_fields(trj.TrajectoryRUMD, read_fields=['type', 'x', 'y', 'z', 'vx', 'vy', 'vz'], ignore=['mass'])
        # This must fail: writing velocities but not reading them
        # self._read_write_fields(trj.TrajectoryRUMD, read_fields=['species', 'position'], ignore=['mass'], fail=True)
        # self._read_write_fields(trj.TrajectoryRUMD, read_fields=['type', 'x', 'y', 'z'], ignore=['mass'], fail=True)
        # TODO: add write_system() to supertrajectory
        # self._read_write(trj.SuperTrajectoryRUMD, self.inpdir, ignore=['id', 'name'])

    def test_pdb(self):
        self._read_write(trj.TrajectoryPDB, ignore=['mass', 'velocity'])
        with trj.TrajectoryPDB('data/trajectory.pdb') as th:
            self.assertTrue(numpy.all(th[0].cell.side == numpy.array([10.0, 10.0, 10.0])))
            self.assertTrue(numpy.all(th[1].cell.side == numpy.array([10.0, 10.0, 10.0])))

    def test_csv(self):
        import atooms.trajectory.csv
        self._read_write(atooms.trajectory.csv.TrajectoryCSV, ignore=['mass', 'species', 'velocity'])
        self._append(atooms.trajectory.csv.TrajectoryCSV, ignore=['mass', 'species', 'velocity'])

    def test_dynamo(self):
        import glob
        from atooms.trajectory import SuperTrajectory, TrajectoryXYZ, TrajectoryDynamO

        with trj.TrajectoryDynamO('data/dynamo/one/config.start.xml') as th:
            self.assertEqual(th.steps, [0])
            self.assertAlmostEqual(th[0].density, 0.5)

        for path in ['data/dynamo/one', 'data/dynamo/two']:
            with trj.TrajectoryDynamO(path) as th:
                th_xyz = th.copy(TrajectoryXYZ,
                                 fout=self.inpdir + '/conv.xyz',
                                 only=['species', 'position'])
                th_xyz.close()
                with TrajectoryXYZ(self.inpdir + '/conv.xyz') as th_xyz:
                    self.assertEqual(th.steps, th_xyz.steps)
                    self.assertEqual(str(th[0]), str(th_xyz[0]))
                    self.assertEqual(str(th[-1]), str(th_xyz[-1]))

    def test_super(self):
        import glob
        with TrajectoryXYZ(os.path.join(self.inpdir, '0.xyz'), 'w') as th:
            th.timestep = 0.001
            th.write(self.system[0], 10)
            th.write(self.system[0], 20)
        with TrajectoryXYZ(os.path.join(self.inpdir, '1.xyz'), 'w') as th:
            th.timestep = 0.001
            th.write(self.system[0], 30)
            th.write(self.system[0], 40)
        with trj.SuperTrajectory(glob.glob(self.inpdir + '/*'), TrajectoryXYZ) as th:
            self.assertFalse(th.grandcanonical)
            self.assertEqual(th.times, [0.001*10, 0.001*20, 0.001*30, 0.001*40])
            self.assertEqual(th.timestep, 0.001)
            self.assertEqual(th.steps, [10, 20, 30, 40])
            for i, system in enumerate(th):
                self.assertTrue(_equal(self.system[0], system, ignore=['mass', 'velocity']))

    def test_super_rumd(self):
        # TODO: refactor supertrajectory tests
        # RUMD uses integer ids for chemical species. They should be integers.
        for s in self.system:
            s.particle = _rename_species(s.particle, {'A': '0', 'B': '1'})
        with trj.TrajectoryRUMD(os.path.join(self.inpdir, '0.xyz.gz'), 'w') as th:
            th.timestep = 0.001
            th.write(self.system[0], 0)
        with trj.TrajectoryRUMD(os.path.join(self.inpdir, '1.xyz.gz'), 'w') as th:
            th.timestep = 0.001
            th.write(self.system[0], 1)
        with trj.rumd.SuperTrajectoryRUMD(self.inpdir) as th:
            self.assertEqual(th.times, [0.001*0, 0.001*1])
            self.assertEqual(th.timestep, 0.001)
            self.assertEqual(th.steps, [0, 1])
            for i, system in enumerate(th):
                self.assertTrue(_equal(self.system[i], system, ignore=['mass']))

    @unittest.expectedFailure
    def test_super_rumd_fields(self):
        for s in self.system:
            s.particle = _rename_species(s.particle, {'A': '0', 'B': '1'})
        with trj.TrajectoryRUMD(os.path.join(self.inpdir, '0.xyz.gz'), 'w') as th:
            th.timestep = 0.001
            th.write(self.system[0], 0)
        with trj.TrajectoryRUMD(os.path.join(self.inpdir, '1.xyz.gz'), 'w') as th:
            th.timestep = 0.001
            th.write(self.system[0], 1)
        # TODO: supertrajectories should propagate fields but currently do not
        with trj.rumd.SuperTrajectoryRUMD(self.inpdir) as th:
            th.variables = ['type', 'x', 'y', 'z']
            system = th[0]
            self.assertTrue(_equal(self.system[0], system, ignore=['mass']), fail=True)
            self.assertTrue(self.system[0].__class__ is system.__class__)

    def test_folder(self):
        import glob
        with TrajectoryXYZ(os.path.join(self.inpdir, '10.xyz'), 'w') as th:
            th.timestep = 0.001
            th.write(self.system[0], 10)
        with TrajectoryXYZ(os.path.join(self.inpdir, '20.xyz'), 'w') as th:
            th.timestep = 0.001
            th.write(self.system[0], 20)
        with trj.folder.Foldered(self.inpdir, cls='xyz') as th:
            self.assertEqual(th.times, [0.001*10, 0.001*20])
            self.assertEqual(th.timestep, 0.001)
            self.assertEqual(th.steps, [10, 20])
            for i, system in enumerate(th):
                self.assertTrue(_equal(self.system[i], system, ignore=['mass', 'velocity']))

    def test_cache(self):
        with TrajectoryXYZ(os.path.join(self.inpdir, 'cache.xyz'), 'w') as th:
            for i in range(10):
                th.write(self.system[0], i)

        with TrajectoryXYZ(os.path.join(self.inpdir, 'cache.xyz'), 'r') as th_0:
            s_0 = th_0[0]

        with TrajectoryXYZ(os.path.join(self.inpdir, 'cache.xyz'), 'r') as th_1:
            th_1.cache = True
            s_1 = th_1.read(0)
            s_1 = th_1.read(0)

    def test_block_size(self):
        from atooms.trajectory.utils import check_block_size, get_block_size
        steps = [2**i for i in range(5)]
        block = get_block_size(steps)
        check_block_size(steps, block)

    def test_decorator(self):
        """Test that frame is accessible to callbacks"""
        finp = os.path.join(self.inpdir, 'test.xyz')
        with open(finp, 'w') as fh:
            fh.write("""\
2
columns:id,x,y,z step:1 cell:5.0,5.0,5.0
B 1.0 -1.0 0.0
A 2.9 -2.9 0.0
2
columns:id,x,y,z step:2 cell:5.0,5.0,5.0
C 1.0 -1.0 0.0
B 2.9 -2.9 0.0
""")

        def cbk(system):
            system.frame
            return system
        with TrajectoryXYZ(finp) as th:
            th.add_callback(cbk)
            th[0]
            th[1]

    def _copy_inplace(self, trajectory, expect=False):
        """
        Test that trajectory returns a copy of the system and that
        modifications are not propagated to the underlying trajectory.

        Test in-place modification
        """
        system = trajectory[0]
        original = system.particle[1].position.copy()
        system.particle[1].position *= 2
        new_system = trajectory[0]
        if expect:
            self.assertEqual(system.particle[1].position[0], new_system.particle[1].position[0])
        else:
            self.assertNotEqual(system.particle[1].position[0], new_system.particle[1].position[0])

    def _copy_reassign(self, trajectory, expect=False):
        """
        Test that trajectory returns a copy of the system and that
        modifications are not propagated to the underlying trajectory.

        Test assignement
        """
        system = trajectory[0]
        original = system.particle[1].position.copy()
        system.particle[1].position[0] = 100000000000000.0
        if expect:
            self.assertEqual(system.particle[1].position[0], new_system.particle[1].position[0])
        else:
            self.assertNotEqual(system.particle[1].position[0], new_system.particle[1].position[0])

    def test_copy_ram_view(self):
        with trj.ram.TrajectoryRamView() as th:
            th[0] = self.system[0]
            self._copy_inplace(th, expect=True)

    def test_copy_ram(self):
        with trj.ram.TrajectoryRam() as th:
            th[0] = self.system[0]
            self._copy_inplace(th)

    def test_copy_xyz(self):
        with trj.TrajectoryXYZ(self.inpfile, 'w') as th:
            th.write(self.system[0])
        with trj.TrajectoryXYZ(self.inpfile, 'r') as th:
            self._copy_inplace(th)

    def test_info(self):
        with trj.xyz.TrajectoryXYZ(self.inpfile, 'w') as th:
            th.write(self.system[0])
            th.write(self.system[1])
        with trj.xyz.TrajectoryXYZ(self.inpfile, 'r') as th:
            from atooms.trajectory.utils import info
            info(th)
            keys = [
                'path',
                'format',
                'frames',
                'megabytes',
                'particles',
                'species',
                'composition',
                'cell density',
                'cell side',
                'cell volume',
                'steps',
                'duration',
                'timestep',
                'block size',
                'steps between frames',
                'time between frames',
                'block steps',
                'block',
                'grandcanonical']
            info(th, keys=','.join(keys))

    def test_lammps(self):
        import sys
        with open(self.inpfile, 'w') as fh:
            fh.write("""\
ITEM: TIMESTEP
0
ITEM: NUMBER OF ATOMS
2
ITEM: BOX BOUNDS pp pp pp
-3 3
-3 3
-3 3
ITEM: ATOMS id type xs ys zs mass vx vy vz
2 2 0.10 0.11 0.12 3.00 1.0 1.0 1.0
1 1 0.20 0.21 0.22 2.00 1.0 1.0 1.0
ITEM: TIMESTEP
1
ITEM: NUMBER OF ATOMS
2
ITEM: BOX BOUNDS pp pp pp
-4 4
-4 4
-4 4
ITEM: ATOMS id type xs ys zs mass
1 1 0.00 0.01 0.02 2.00
2 2 0.50 0.51 0.52 3.00
""")

        def scale(pos, side):
            return [(x - 0.5) * L for x, L in zip(pos, side)]
        with trj.lammps.TrajectoryLAMMPS(self.inpfile) as th:
            self.assertEqual(list(th[0].cell.side), [6.0, 6.0, 6.0])
            self.assertEqual(list(th[0].particle[0].position), scale([0.20, 0.21, 0.22], [6.0, 6.0, 6.0]))
            self.assertEqual(list(th[0].particle[1].position), scale([0.10, 0.11, 0.12], [6.0, 6.0, 6.0]))
            self.assertEqual(list(th[0].particle[0].velocity), [1.0, 1.0, 1.0])
            self.assertEqual(list(th[1].cell.side), [8.0, 8.0, 8.0])
            self.assertEqual(list(th[1].particle[0].position), scale([0.00, 0.01, 0.02], [8.0, 8.0, 8.0]))
            self.assertEqual(list(th[1].particle[1].position), scale([0.50, 0.51, 0.52], [8.0, 8.0, 8.0]))
            self.assertEqual(th[0].particle[0].mass, 2.0)
            self.assertEqual(th[0].particle[1].mass, 3.0)
            system = th[0]

        with trj.lammps.TrajectoryLAMMPS(self.inpfile, 'w') as th:
            th.write(system)

    def test_lammps_style(self):
        with trj.lammps.TrajectoryLAMMPS(self.inpfile, 'w') as th:
            th.write(self.system[0])
        for p in self.system[0].particle:
            p.charge = 1
        with trj.lammps.TrajectoryLAMMPS(self.inpfile, 'w') as th:
            th.write(self.system[0])
        # print(open('/tmp/test_trajectory.inp').read())

    def test_lammps_folder(self):
        with open(self.inpdir + '/0.atom', 'w') as fh:
            fh.write("""\
ITEM: TIMESTEP
10
ITEM: NUMBER OF ATOMS
2
ITEM: BOX BOUNDS pp pp pp
-3 3
-3 3
-3 3
ITEM: ATOMS id type xs ys zs
2 1 0.10 0.11 0.12
1 1 0.20 0.21 0.22
""")
        with open(self.inpdir + '/1.atom', 'w') as fh:
            fh.write("""\
ITEM: TIMESTEP
20
ITEM: NUMBER OF ATOMS
2
ITEM: BOX BOUNDS pp pp pp
-4 4
-4 4
-4 4
ITEM: ATOMS id type xs ys zs
1 1 0.00 0.01 0.02
2 1 0.50 0.51 0.52
""")
        from atooms.trajectory import TrajectoryFolderLAMMPS

        def scale(pos, side):
            return [(x - 0.5) * L for x, L in zip(pos, side)]
        with TrajectoryFolderLAMMPS(self.inpdir) as th:
            self.assertEqual(th.steps, [10, 20])
            self.assertEqual(list(th[0].cell.side), [6.0, 6.0, 6.0])
            self.assertEqual(list(th[0].particle[0].position), scale([0.20, 0.21, 0.22], [6.0, 6.0, 6.0]))
            self.assertEqual(list(th[0].particle[1].position), scale([0.10, 0.11, 0.12], [6.0, 6.0, 6.0]))
            self.assertEqual(list(th[1].cell.side), [8.0, 8.0, 8.0])
            self.assertEqual(list(th[1].particle[0].position), scale([0.00, 0.01, 0.02], [8.0, 8.0, 8.0]))
            self.assertEqual(list(th[1].particle[1].position), scale([0.50, 0.51, 0.52], [8.0, 8.0, 8.0]))

    def test_lammps_triclinic(self):
        import sys
        with open(self.inpfile, 'w') as fh:
            fh.write("""\
ITEM: TIMESTEP
0
ITEM: NUMBER OF ATOMS
2
ITEM: BOX BOUNDS xy xz yz pp pp pp
1 20 0.0000000000000000e+00
1 20 0.0000000000000000e+00
1 20 0.0000000000000000e+00
ITEM: ATOMS id type x y z q
1 1 12.6334 21.0623 23.5327 1
2 1 17.2724 17.2329 21.5826 1
""")
        with trj.lammps.TrajectoryLAMMPS(self.inpfile) as th:
            self.assertAlmostEqual(list(th[0].cell.side), [19, 19, 19])
            self.assertEqual(sorted(th.variables), ['particle.charge',
                                                    'particle.position',
                                                    'particle.species'])

    def test_lammps_images(self):
        import sys
        with open(self.inpfile, 'w') as fh:
            fh.write("""\
ITEM: TIMESTEP
0
ITEM: NUMBER OF ATOMS
2
ITEM: BOX BOUNDS xy xz yz pp pp pp
1 20 0.0000000000000000e+00
1 20 0.0000000000000000e+00
1 20 0.0000000000000000e+00
ITEM: ATOMS id type x y z ix iy iz
1 1 12.6334 21.0623 23.5327 0 0 1
2 1 17.2724 17.2329 21.5826 0 0 1
""")
        with trj.lammps.TrajectoryLAMMPS(self.inpfile) as th:
            s = th[0]
            self.assertEqual(s.particle[0]._ix, 0)
            self.assertEqual(s.particle[0]._iz, 1)
            self.assertEqual(s.particle[1]._iz, 1)
        with trj.lammps.TrajectoryLAMMPS(self.inpfile, 'w') as th:
            th.write(s)

    def test_xyz_gz(self):
        inp = os.path.join(os.path.dirname(__file__),
                           '../data/ka_rumd_N1200.xyz.gz')
        with trj.TrajectoryXYZ(inp) as th:
            self.assertEqual(len(th), 17)

    def test_split(self):
        with trj.TrajectoryXYZ(os.path.join(self.inpdir, 'test.xyz'), 'w') as th:
            th.write(self.system[0])
            th.write(self.system[0])
            fname = th.filename
        from atooms.trajectory.utils import split, field
        with trj.TrajectoryXYZ(fname) as th:
            x = field(th, th, 'position', 0)
            split(th, archive=True)
        import glob
        self.assertTrue(os.path.exists(fname + '.tar.gz'))

    def test_slice_twice(self):
        from atooms.trajectory import Sliced
        import glob
        with trj.TrajectoryRUMD(os.path.join(self.inpdir, 'trajectory0.xyz'), 'w') as th:
            th.timestep = 0.001
            th.write(self.system[0], 10)
            th.write(self.system[0], 20)
        with trj.TrajectoryRUMD(os.path.join(self.inpdir, 'trajectory1.xyz'), 'w') as th:
            th.timestep = 0.001
            th.write(self.system[0], 30)
            th.write(self.system[0], 40)
        with trj.SuperTrajectoryRUMD(self.inpdir, basename="trajectory*") as th:
            with Sliced(th, slice(None, None, 1)) as ts:
                for s in ts:
                    pass
            with Sliced(th, slice(3, None, 1)) as ts:
                for s in ts:
                    pass

    def test_array_formatting(self):
        variables = ['position', 'species']

        def _list(x):
            try:
                return [_ for _ in x]
            except:
                return [x]
        from itertools import chain
        fmt = ' '.join(['{0.' + variable.split('particle.')[-1] + '}' for variable in variables]) + '\n'
        for p in self.system[0].particle:
            # a = replace(']', '').strip()
            a = [_list(getattr(p, variable)) for variable in variables]
            # print(type(a[0][0]), type(a[1]))
            # print(list(chain(a)))

    def test_callbacks_decorated_class(self):
        """
        Check that callbacks added to decorated classes do not propagate
        to parent and that Sliced() works well when cache is enabled
        in the parent trajectory
        """
        from atooms.trajectory import Sliced, Unfolded
        from atooms.trajectory.decorators import fold

        # Check Unfolded
        th = trj.TrajectoryXYZ(self.inpfile, 'w')
        th.write(self.system[0])
        th.write(self.system[1])
        th.close()
        th = trj.TrajectoryXYZ(self.inpfile)
        ts = Unfolded(th)
        ts.add_callback(fold)
        self.assertTrue(len(th.callbacks) == 0)

        # Check Sliced
        th = trj.TrajectoryRam()
        th.write(self.system[0])
        ts = Sliced(th, slice(None, None, None))
        ts.add_callback(fold)
        self.assertTrue(len(th.callbacks) == 0)

        # Check Sliced with caching
        th = trj.TrajectoryXYZ(self.inpfile, 'w')
        th.write(self.system[0])
        self.system[1].particle[0].position += 0.1
        th.write(self.system[1])
        th.close()
        th = trj.TrajectoryXYZ(self.inpfile)
        th.cache = True
        s0 = th[0]
        s1 = th[1]
        ts = Sliced(th, slice(1, None, 1))
        s1b = ts[0]
        self.assertTrue(numpy.all(s1.particle[0].position == s1b.particle[0].position))

    def tearDown(self):
        rmf(self.inpfile)
        rmd(self.inpdir)


if __name__ == '__main__':
    unittest.main()



import numpy

__all__ = ['report_simulation', 'report_stats', 'report_system',
           'report_trajectory']


def report_trajectory(trajectory):
    """Return a dict with information about a `trajectory` instance."""
    db = {
        'frames': len(trajectory),
        'steps': trajectory.steps[-1],
        'duration': trajectory.times[-1]
    }
    if len(trajectory) > 1:
        db['timestep'] = trajectory.timestep
        db['grandcanonical'] = trajectory.grandcanonical
        db['block size'] = trajectory.block_size
        if trajectory.block_size == 1:
            db['steps_between_frames'] = trajectory.steps[1]-trajectory.steps[0]
            db['time_between_frames'] = trajectory.times[1]-trajectory.times[0]
        else:
            db['block steps'] = trajectory.steps[trajectory.block_size-1]
            db['block'] = [trajectory.steps[i] for i in range(trajectory.block_size)]
    return db

def report_system(system):
    """Return a dict with information about a `system` instance."""
    db = {
        'number_of_particles': len(system.particle),
        'number_of_species': len(system.distinct_species),
        'species': [numpy.asarray(_).item() for _ in system.distinct_species],
        'composition': [system.composition[numpy.asarray(_).item()] for _ in system.distinct_species],
        'concentration': [system.concentration[numpy.asarray(_).item()] for _ in system.distinct_species],
        'size_dispersion': float(numpy.std([p.radius for p in system.particle]) / numpy.mean([p.radius for p in system.particle])),
        'density': float(round(system.density, 10)),
        'cm_velocity_norm': float(numpy.sum(system.center_of_mass('velocity')**2)**0.5)
    }
    if system.cell is not None:
        db['cell_side'] = [_.item() for _ in system.cell.side]
        db['cell_volume'] = system.cell.volume.item()
    return db

def report_simulation(simulation):
    """Return a dict with information about a `simulation` instance."""
    # TODO: the simulation instance should provide its params as dict
    db = {
        'backend': str(simulation.backend.__class__),
        'rmsd': simulation.rmsd.item() if hasattr(simulation.rmsd, 'item') else simulation.rmsd,
        'steps': simulation.current_step,
        'wall_time': simulation.wall_time(),
        'TSP': simulation.wall_time(per_step=True, per_particle=True)
    }
    if hasattr(simulation.backend, 'timestep'):
        db['timestep'] = simulation.backend.timestep
    if hasattr(simulation.backend, 'partial_rmsd'):
        db['partial_rmsd'] = simulation.backend.partial_rmsd
    return db

def report_stats(data, ignore=('steps',)):
    """Return a dict with stats about a `data` dictionary with time series."""
    db = {}
    keys = [key for key in data if key not in ignore]
    for key in keys:
        key.replace(' ', '_')
        db[f'mean_{key}'] = numpy.mean(data[key]).item()
    for key in keys:
        key.replace(' ', '_')
        db[f'std_{key}'] = numpy.std(data[key]).item()
    return db



"""Linked cells to compute neighbors."""

from collections import defaultdict
import numpy


class LinkedCells:

    def __init__(self, rcut, newton=False):
        self.rcut = rcut
        self.neighbors = []
        self.newton = newton
        self._is_adjusted = False

    def _adjust(self, box, periodic=None):
        self.box = box
        self.hbox = box / 2
        self.n_cell = (box / self.rcut).astype(int)
        self.box_cell = self.box / self.n_cell
        # if periodic is None:
        periodic = numpy.array([True] * len(box))
        self._map(self.newton, periodic)

    def _map(self, newton, periodic):
        def _pbc(t, N):
            """Apply PBCs to cell only along periodic directions"""
            for i in range(len(t)):
                if not periodic[i]:
                    continue
                if t[i] >= N[i]:
                    t[i] -= N[i]
                elif t[i] < 0:
                    t[i] += N[i]
            return t

        def _outside(t, N):
            """Check whether subcell is outside cell"""
            for i in range(len(t)):
                if t[i] >= N[i] or t[i] < 0:
                    return True
            return False

        def _map_3d_newton(n_cell):
            neigh_cell = {}
            for ix in range(n_cell[0]):
                for iy in range(n_cell[1]):
                    for iz in range(n_cell[2]):
                        # This is a map of neighbouring cells obeying III law
                        neigh_cell[(ix, iy, iz)] = \
                            [(ix+1, iy, iz), (ix+1, iy+1, iz), (ix, iy+1, iz),
                             (ix-1, iy+1, iz), (ix+1, iy, iz-1), (ix+1, iy+1, iz-1),
                             (ix, iy+1, iz-1), (ix-1, iy+1, iz-1), (ix+1, iy, iz+1),
                             (ix+1, iy+1, iz+1), (ix, iy+1, iz+1), (ix-1, iy+1, iz+1),
                             (ix, iy, iz+1)]
            return neigh_cell

        def _map_2d_newton(n_cell):
            neigh_cell = {}
            for ix in range(n_cell[0]):
                for iy in range(n_cell[1]):
                    # This is a map of neighbouring cells obeying III law
                    neigh_cell[(ix, iy)] = \
                        [(ix+1, iy), (ix+1, iy+1), (ix, iy+1), (ix-1, iy+1)]
            return neigh_cell

        def _map_3d_nonewton(n_cell):
            neigh_cell = {}
            for ix in range(n_cell[0]):
                for iy in range(n_cell[1]):
                    for iz in range(n_cell[2]):
                        neigh_cell[(ix, iy, iz)] = []
                        for deltax in [-1, 0, 1]:
                            for deltay in [-1, 0, 1]:
                                for deltaz in [-1, 0, 1]:
                                    if deltax == deltay == deltaz == 0:
                                        continue
                                    neigh_cell[(ix, iy, iz)].append((ix+deltax, iy+deltay, iz+deltaz))
            return neigh_cell

        def _map_2d_nonewton(n_cell):
            neigh_cell = {}
            for ix in range(n_cell[0]):
                for iy in range(n_cell[1]):
                    neigh_cell[(ix, iy)] = []
                    for deltax in [-1, 0, 1]:
                        for deltay in [-1, 0, 1]:
                            if deltax == deltay == 0:
                                continue
                            neigh_cell[(ix, iy)].append((ix+deltax, iy+deltay))
            return neigh_cell

        if len(self.n_cell) == 3:
            if newton:
                self._neigh_cell = _map_3d_newton(self.n_cell)
            else:
                self._neigh_cell = _map_3d_nonewton(self.n_cell)
        elif len(self.n_cell) == 2:
            if newton:
                self._neigh_cell = _map_2d_newton(self.n_cell)
            else:
                self._neigh_cell = _map_2d_nonewton(self.n_cell)
        else:
            raise ValueError('linked cells not supported for dimensions not in {2,3}')

        # Apply PBC
        for key in self._neigh_cell:
            for idx in range(len(self._neigh_cell[key])):
                folded = _pbc(list(self._neigh_cell[key][idx]), self.n_cell)
                self._neigh_cell[key][idx] = tuple(folded)

        # Remove subcells that are out of bounds
        # (this is only applied to non periodic directions)
        new_neigh_cell = {}
        for key in self._neigh_cell:
            new_neigh_cell[key] = []
            for subcell in self._neigh_cell[key]:
                if not _outside(subcell, self.n_cell):
                    new_neigh_cell[key].append(subcell)
        self._neigh_cell = new_neigh_cell

    def _index(self, pos):
        # TODO: add dimension to hbox
        x = ((pos.transpose() + self.hbox) / self.box_cell)
        return x.astype(numpy.int32)

    def compute(self, box, pos, ids, newton=True):
        if not self._is_adjusted:
            self._adjust(box, periodic=None)
            self._is_adjusted = True

        self.neighbors = []
        self.number_neighbors = []
        index = self._index(pos)
        particle_in_cell = defaultdict(list)
        for ipart, icell in enumerate(index):
            particle_in_cell[tuple(icell)].append(ipart)

        for ipart in range(pos.shape[1]):
            icell = tuple(index[ipart])
            # Initialize an empty list
            neighbors = []
            # Start with particles in the cell of particle ipart
            if self.newton:
                neighbors += [_ for _ in particle_in_cell[icell] if _ > ipart]
            else:
                neighbors += [_ for _ in particle_in_cell[icell] if _ != ipart]
            # Loop over neighbors cells and add neighbors
            for jcell in self._neigh_cell[icell]:
                neighbors += particle_in_cell[jcell]
            self.neighbors.append(neighbors)
            self.number_neighbors.append(len(neighbors))

        npart = len(self.neighbors)
        number_neighbors = numpy.array(self.number_neighbors)
        # neighbors_array = numpy.ndarray((npart, max(number_neighbors)), dtype=numpy.int32)
        neighbors_array = numpy.ndarray((max(number_neighbors), npart), dtype=numpy.int32)
        for ipart in range(len(self.neighbors)):
            neighbors_array[0:len(self.neighbors[ipart]), ipart] = self.neighbors[ipart]
        self.neighbors = neighbors_array
        self.number_neighbors = number_neighbors



# This file is part of atooms
# Copyright 2010-2024, Daniele Coslovich

"""
Simulation backends.

A backend is an adapter to an actual implementation of a
simulation. The simulation backend interface must provide, at a
minimum, a `run()` method that performs the simulation. See the
`DryRun` backend for the interface that a valid backend should
implement.
"""



# This file is part of atooms
# Copyright 2010-2024, Daniele Coslovich

"""
Optimization framework for systems of interacting particles.

`atooms` provides a generic optimization interface that abstracts out
most of the common parts of particle-based optimizations.
"""

from .core import Optimization



#!/usr/bin/env python

import os
import unittest

from atooms import trajectory


class Test(unittest.TestCase):

    def setUp(self):
        self.finp = '/tmp/test_hoomd.xyz'
        fh = open(self.finp, 'w')
        fh.write("""\
2
cell:6.0,6.0,6.0
A 1.0 -1.0 0.0
A 2.9 -2.9 0.0
2
cell:6.0,6.0,6.0
A 1.1 -1.1 0.0
A -2.9 -2.9 0.0
2
cell:6.0,6.0,6.0
A 1.2 -1.2 0.0
A -2.9 2.9 0.0
2
cell:6.0,6.0,6.0
A 1.3 -1.3 0.0
A -2.8 2.8 0.0
""")
        fh.close()

    def test(self):
        fname = '/tmp/test_hoomd.tgz'
        txyz = trajectory.TrajectoryXYZ(self.finp)

        with trajectory.TrajectoryHOOMD(fname, 'w:gz') as t:
            t.write(txyz[0], 0)

        with trajectory.TrajectoryHOOMD(fname, 'r') as t:
            self.assertEqual(t[0].particle[0].position[1],
                             txyz[0].particle[0].position[1])

            self.assertEqual(t[0].cell.side[1], txyz[0].cell.side[1])

        txyz.close()

    def tearDown(self):
        os.remove('/tmp/test_hoomd.tgz')
        os.remove('/tmp/test_hoomd.xyz')


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

import sys
import os
import unittest
import numpy as np

try:
    from atooms.trajectory.gsd import TrajectoryGSD
    HAS_GSD = True
except ImportError:
    HAS_GSD = False


class Test(unittest.TestCase):

    def setUp(self):
        if sys.version_info.major == 2:
            self.skipTest('cannot test GSD with python 2')
        if not HAS_GSD:
            self.skipTest('cannot load GSD format (missing hoomd?)')
        self.fname3d = os.path.join(os.path.dirname(__file__), '../data/trajectory3d.gsd')
        self.fname2d = os.path.join(os.path.dirname(__file__), '../data/trajectory2d.gsd')

    def test(self):
        fname = self.fname3d
        traj = TrajectoryGSD(fname)
        p = traj[0].particle[0]

        # Sanity checks.
        self.assertEqual(len(traj), 4)
        self.assertEqual(traj.steps, [0, 1, 2, 3])
        self.assertEqual(traj[0].distinct_species, ['A', 'B'])
        self.assertTrue(np.all(np.isclose(traj[0].cell.side, np.array([10.68387318, 10.68387318, 10.68387318]), 1e-12)))
        self.assertEqual(traj[0].number_of_dimensions, 3)

        self.assertEqual(p.species, 'B')
        self.assertEqual(p.mass, 1.0)
        self.assertEqual(p.radius, 0.5)
        self.assertTrue(np.all(p.position == np.array([5.3419366, 5.3419366, 5.3419366], dtype=np.float32)))

    def test2d(self):
        fname = self.fname2d
        traj = TrajectoryGSD(fname)
        p = traj[0].particle[0]

        # Sanity checks.
        self.assertEqual(len(traj), 4)
        self.assertEqual(traj.steps, [0, 1, 2, 3])
        self.assertEqual(traj[0].distinct_species, ['A', 'B'])
        self.assertTrue(np.all(np.isclose(traj[0].cell.side, np.array([10.78327751, 10.78327751]), 1e-12)))
        self.assertEqual(traj[0].number_of_dimensions, 2)

        self.assertEqual(p.species, 'A')
        self.assertEqual(p.mass, 1.0)
        self.assertEqual(p.radius, 0.5)
        self.assertTrue(np.all(p.position == np.array([5.3916388, 5.3916388], dtype=np.float32)))

    def tearDown(self):
        pass


if __name__ == '__main__':
    unittest.main()



"""Helper functions"""

import os
import sys
import shutil
import time
import logging
import warnings


# Parallel environment

try:
    # pylint:disable=import-error
    from mpi4py import MPI
    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()
    size = comm.Get_size()
    sys_excepthook = sys.excepthook

    def mpi_excepthook(v, t, tb):
        sys_excepthook(v, t, tb)
        MPI.COMM_WORLD.Abort(1)

    sys.excepthook = mpi_excepthook
except:
    comm = None
    rank = 0
    size = 1


def barrier():
    """Syncronize processes in parallel."""
    if size > 1:
        comm.barrier()


# Utility functions to mimic bash directory / file handling

def mkdir(dirname):
    """
    Create a directory `dirname` or a list `dirname` of directories,
    silently ignoring existing directories.

    This is just a wrapper to `os.makedirs`. All intermediate
    subdirectories are created as needed.
    """
    if dirname is None:
        return
    if isinstance(dirname, str):
        dirs = [dirname]
    else:
        dirs = dirname

    for dd in dirs:
        try:
            os.makedirs(dd)
        except OSError:
            pass


def rmd(files):
    """Totally silent wrapper to shutil.rmtree."""
    try:
        shutil.rmtree(files)
    except:
        pass


def rmf(files):
    """
    Remove `files` without complaining.

    The variable `files` can be a list or tuple of paths or a single
    string parseable by glob.glob().
    """
    import glob
    try:
        # This a single pattern
        for pathname in glob.glob(files):
            try:
                os.remove(pathname)
            except OSError:
                # File does not exists or it is a folder
                pass
    except (TypeError, AttributeError):
        # This is a list
        for pathname in files:
            try:
                os.remove(pathname)
            except OSError:
                # File does not exists or it is a folder
                pass


def cp(finp, fout):
    """
    Copy `finp` to `fout`.

    Wrapper to shutil.copy().
    """
    # Avoid overwriting file
    if finp == fout:
        return
    shutil.copy(finp, fout)


def wget(url, output_dir):
    """
    Python implementation of bash wget
    """
    try:
        from urllib.request import urlopen  # Python 3
    except ImportError:
        from urllib2 import urlopen  # Python 2

    basename = os.path.basename(url)
    output_file = os.path.join(output_dir, basename)
    response = urlopen(url)
    length = 16*1024
    with open(output_file, 'wb') as fh:
        shutil.copyfileobj(response, fh, length)


# Alias of wget
download = wget


# Timings

class Timer:

    """Timer class inspired by John Paulett's stopwatch class."""

    def __init__(self, iterations=None, output=sys.stdout):
        self.__start_cpu = None
        self.__start_wall = None
        self.output = output
        self.cpu_time = 0.0
        self.wall_time = 0.0
        self.iterations = iterations
        try:
            self._wall_time_func = MPI.Wtime
        except:
            self._wall_time_func = time.time

    def __str__(self):
        return 'timer wall time [s]: {:.2f}, cpu time [s]: {:.2f}'.format(self.wall_time, self.cpu_time)

    def __repr__(self):
        return 'timer wall time [s]: {:.2f}, cpu time [s]: {:.2f}'.format(self.wall_time, self.cpu_time)

    def start(self):
        self.__start_cpu = self.__now_cpu()
        self.__start_wall = self.__now_wall()

    def stop(self):
        if self.__start_cpu is None:
            raise ValueError("Timer not started")
        self.cpu_time += self.__now_cpu() - self.__start_cpu
        self.wall_time += self.__now_wall() - self.__start_wall

    def __now_cpu(self):
        try:
            return time.clock()
        except AttributeError:
            return time.process_time()

    def __now_wall(self):
        return self._wall_time_func()

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *args):
        self.stop()
        if self.output is not None:
            print(self, file=self.output)


def clockit(func, output=sys.stdout):
    """
    Function decorator that times the evaluation of `func` and prints
    the execution time.
    """
    def new(*args, **kw):
        t = Timer()
        t.start()
        retval = func(*args, **kw)
        t.stop()
        print('%s in %s' % (func.__name__, t), file=output)
        del t
        return retval
    return new


def fractional_slice(first, last, skip, n):
    """
    Return a slice assuming `first` or `last` are fractions of `n`,
    the length of the iterable, if `first` or `last` are in (0,1)
    """
    # We use an implicit convention here:
    # If first or last are in (0,1) then they are considered as fractions of the iterable
    # otherwise they are integer indexes. Note the explicit int() cast in the latter case.
    if first is not None:
        if 0 < first < 1:
            first = int(first * n)
        else:
            first = int(first)

    if last is not None:
        if 0 < last < 1:
            last = int(last * n)
        else:
            last = int(last)

    return slice(first, last, skip)


def add_first_last_skip(parser, what=None):
    """
    Add first, last, skip arguments to ArgumentParser object.

    Compatible with fractional_slice(). Convenience function for
    analysis scripts.
    """
    if what is None:
        what = ['first', 'last', 'skip']
    if 'first' in what:
        parser.add_argument('-f', '--first', dest='first', type=float, default=None, help='first cfg (accepts fractions)')
    if 'last' in what:
        parser.add_argument('-l', '--last', dest='last', type=float, default=None, help='last cfg (accepts fractions)')
    if 'skip' in what:
        parser.add_argument('-s', '--skip', dest='skip', type=int, default=1, help='interval between cfg')
    return parser


# Logging facilities

def log_to_stderr(level=None):
    """Turn on logging and add a handler which prints to stderr"""
    warnings.warn('log_to_stderr() is deprecated', DeprecationWarning)


class _ParallelFilter(logging.Filter):
    def filter(self, rec):
        if hasattr(rec, 'rank'):
            if rec.rank == 'all':
                return True
            else:
                return rank == rec.rank
        else:
            return rank == 0


class _MyFormatter(logging.Formatter):
    def format(self, record):
        if record.levelname in ['WARNING', 'ERROR']:
            return '# ' + record.levelname + ' ' + record.msg % record.args
        return '# ' + record.msg % record.args


def setup_logging(name=None, level=40, filename=None, update=False):
    """Logging API."""
    if name is None:
        log = logging.getLogger()
    else:
        log = logging.getLogger(name)

    if update:
        # We only update the level of the logger
        log.setLevel(level)
    else:
        # The logger should always pass messages to all handlers
        current_level = log.getEffectiveLevel()
        log.setLevel(min(level, current_level))

    formatter = _MyFormatter()
    if filename is None:
        handler = logging.StreamHandler(sys.stdout)
    else:
        handler = logging.FileHandler(filename)

    handler.setFormatter(formatter)
    # From the doc: "Note that filters attached to handlers are
    # consulted before an event is emitted by the handler, whereas
    # filters attached to loggers are consulted whenever an event is
    # logged (using debug(), info(), etc.), before sending an event to
    # handlers. This means that events which have been generated by
    # descendant loggers will not be filtered by a logger filter
    # setting, unless the filter has also been applied to those
    # descendant loggers."
    handler.addFilter(_ParallelFilter())
    handler.setLevel(level)
    if update:
        if len(log.handlers) == 0:
            log.addHandler(handler)
    else:
        log.addHandler(handler)

    return log

def clear_logging():
    import logging
    log = logging.getLogger()
    for h in log.handlers[-2:]:
        log.removeHandler(h)

def tipify(s):
    """
    Convert a string into the best matching type.

    The only risk is if a variable is required to be float,
    but is passed without dot.

    Examples
    --------
    .. code-block:: python

       assert type(tipify("2.0")) is float
       assert type(tipify("2")) is int
       assert type(tipify("foo")) is str
    """
    if '_' in s:
        return s
    try:
        return int(s)
    except ValueError:
        try:
            return float(s)
        except ValueError:
            return s


# Miscellaneous

def __header_dict(line):
    # Array entry have comma separated elements, split them into lists
    params = {}
    for key, value in [d.split('=') for d in line.split()]:
        params[key] = value
    return params


def report_parameters(params, fileout, version, comment=''):
    """Report parameters."""
    maxlen = max([len(key) for key in params])
    fmt = comment + '%-' + str(maxlen) + 's = %s\n'
    txt = ""
    txt += fmt % ('version', version)
    for key in sorted(params.keys()):
        txt += fmt % (key, params[key])
    if fileout is not None:
        with open(fileout, 'w') as fh:
            fh.write(txt)
    return txt

def report_command(cmd, params, main, fileout):
    """Report command line options."""
    txt = cmd + ' \\\n'
    for key in sorted(params.keys()):
        if key in main:
            continue
        flag = key.replace('_', '-')
        value = params[key]
        if value is None or value is False:
            continue
        if value is True:
            value = ''
        txt += ' --%s %s \\\n' % (flag, value)
    txt += ' '.join([params[key] for key in main])
    if fileout is not None:
        with open(fileout, 'w') as fh:
            fh.write(txt)
    return txt


class OrderedSet:

    """
    Simple class to store an ordered set of items.

    It does not try to reproduce either set or list interface. It just
    provides a simple interface to deal with the set of distinct
    chemical species in a system as it gets populated e.g. when
    reading a trajectory. This covers the use case of grand-canonical
    simulations.

    Examples
    --------

    .. code-block:: python

       particle = [Particle(species='A'), Particle(species='C')]
       periodic_table = OrderedSet()
       periodic_table.update([p.species for p in particle])
       particle = [Particle(species='A'), Particle(species='D')]
       periodic_table.update([p.species for p in particle])
       print(periodic_table.index('C'))
    """

    def __init__(self):
        self.items = []

    def __repr__(self):
        return repr(self.items)

    def __iter__(self):
        return self.items.__iter__()

    def __getitem__(self, key):
        return self.items[key]

    def __setitem__(self, key, item):
        self.items.__setitem__
        return self.items[key]

    def update(self, items):
        sort_needed = False
        for item in items:
            if item not in self.items:
                self.items.append(item)
                sort_needed = True
        if sort_needed:
            self.items.sort()

    def index(self, item):
        try:
            return self.items.index(item)
        except ValueError:
            raise ValueError('item %s not in %s' % (item, self))


def is_array(arg):
    """
    Non-pythonic guess whether argument is an array.

    It will match list, tuple and numpy array but fail with iterators / generators.
    For a full fledged alternative, see https://stackoverflow.com/questions/1055360/how-to-tell-a-variable-is-iterable-but-not-a-string/44328500#44328500
    """
    import numpy
    return isinstance(arg, (numpy.ndarray, list, tuple))


def canonicalize(fields, thesaurus):
    """
    Replace entries in `fields` list with those found in `thesaurus`
    """
    if fields is None:
        return []
    _fields = []
    for field in fields:
        try:
            _fields.append(thesaurus[field])
        except KeyError:
            _fields.append(field)
    return _fields



"""
Visualization functions for particles

Each show() method relies on a specific, optional visualization backend.
"""

# These functions used to belong to system.particle

_palette = [
    "#3c44aa",
    "#f0002f",
    "#f58f20",
    "#940650",
    "#f0e694",
]

def _hex_to_rgb(h):
    h = h.lstrip('#')
    return tuple(int(h[i:i + 2], 16) / 256 for i in (0, 2, 4))


def show_3dmol(particle, cell=None, radius=1.0, palette=None):
    """Visualize particles in cell using 3dmol http://3dmol.csb.pitt.edu/

    :param particle: sequence of particles to show
    :param cell: cell that encloses the particles
    :param radius: particle radius
    :param palette: to color the particles
    """
    from .particle import distinct_species
    import py3Dmol

    if palette is None:
        palette = ["#50514f", "#f25f5c", "#ffe066", "#247ba0", "#70c1b3",
                   "#0cce6b", "#c200fb", "#e2a0ff", "#6622cc", "#119822"]
    colors = {}
    for i, s in enumerate(distinct_species(particle)):
        colors[s] = palette[i]
    view = py3Dmol.view()
    view.setBackgroundColor('white')
    for p in particle:
        view.addSphere({'center': {'x': p.position[0], 'y': p.position[1], 'z': p.position[2]},
                        'radius': radius * p.radius, 'color': colors[p.species]})
    if cell is not None:
        view.addBox({'center': {'x': cell.center[0], 'y': cell.center[1], 'z': cell.center[2]},
                     'dimensions': {'w': cell.side[0], 'h': cell.side[1], 'd': cell.side[2]},
                     'wireframe': True, 'color': "#000000"})
    return view


def show_matplotlib(particle, cell, output_file=None, linewidth=3, alpha=0.3,
                    show=False, now=False, outfile=None, axis=(0, 1)):
    """
    Show a snapshot of the `particle`s in the given `cell`, projected on the two
    given `axis`. A matplotlib `Figure` instance is returned for further
    customization or visualization in notebooks.

    :param particle: sequence of particles to show
    :param cell: cell that encloses the particles
    :param output_file: optional file where image is saved (ignored if `None`)
    :param linewidth: particles' circles
    :param alpha: transparency of particles
    :param show: show image immediately
    :param now: show image immediately
    :param axis: project on `axis` (2d list or tuple)
    :param outfile: name of output file  (ignored if `None`)
    """
    import matplotlib.pyplot as plt
    from .particle import distinct_species
    if now:
        show = True
    if output_file is not None:
        outfile = output_file
    color_db = ['b', 'r', 'y', 'g']
    species = distinct_species(particle)
    fig = plt.figure()
    ax = fig.add_subplot(111, aspect='equal')
    ax.axes.get_xaxis().set_visible(False)
    ax.axes.get_yaxis().set_visible(False)
    if cell is not None:
        ax.set_xlim((-cell.side[axis[0]]/2, cell.side[axis[1]]/2))
        ax.set_ylim((-cell.side[axis[0]]/2, cell.side[axis[1]]/2))
        
    for p in particle:
        pos = [p.position[axis[0]], p.position[axis[1]]]
        c = plt.Circle(pos, p.radius,
                       facecolor='white',
                       edgecolor='black', linewidth=linewidth)
        ax.add_artist(c)
        c = plt.Circle(pos, p.radius,
                       facecolor=color_db[species.index(p.species)],
                       edgecolor=None, alpha=alpha)
        ax.add_artist(c)
    if outfile is not None:
        fig.savefig(outfile, bbox_inches='tight')
    # TODO: show should be True by default
    if show:
        plt.show()
    return fig


def show_ovito(particle, cell, output_file=None, color='species',
               radius=0.5, viewport=None, callback=None, tmpdir=None,
               camera_dir=(0, 1, 0), camera_pos=(0, -10, 0),
               size=(640, 480), zoom=True, perspective=False,
               color_map='viridis', color_normalize=False, outfile=None):
    """
    Render image of particles in cell using ovito. The image is returned for
    visualization in jupyter notebooks.

    :param particle:
    :param cell:
    :param output_file:
    :param color:
    :param radius:
    :param viewport:
    :param callback:
    :param tmpdir:
    :param camera_dir:
    :param camera_pos:
    :param size:
    :param zoom:
    :param perspective:
    :param color_map:
    :param color_normalize:
    :param outfile:
    """
    import os
    from ovito.io import import_file  # pylint: disable=import-error
    from ovito.vis import Viewport, TachyonRenderer  # pylint: disable=no-name-in-module

    import tempfile
    from atooms.core.utils import mkdir

    if output_file is not None:
        outfile = output_file

    # Color coding
    color_attr = [getattr(p, color) for p in particle]
    is_discrete = isinstance(color_attr[0], (str, int))
    # Corresponding color system
    if is_discrete:
        # Discrete attribute
        color_db = sorted(list(set(color_attr)))
        color_db.sort()
        palette = [_hex_to_rgb(c) for c in _palette]
        colors = []
        for p in particle:
            colors.append(palette[color_db.index(getattr(p, color))])
    else:
        # Continuous attribute
        try:
            # Try with matplotlib
            import matplotlib
            import matplotlib.cm
            colormap = matplotlib.cm.get_cmap(color_map)
            if color_normalize:
                norm = matplotlib.colors.Normalize()
                norm.autoscale(color_attr)
                colors = colormap(norm(color_attr))
            else:
                colors = colormap(color_attr)
        except ImportError:
            # Fallback
            if color_normalize:
                c_min, c_max = min(color_attr), max(color_attr)
                if c_min != c_max:
                    colors = [[(c - c_min) / (c_max - c_min), 0.3,
                               (c_max - c) / (c_max - c_min)] for c in color_attr]
            else:
                colors = [[c, 0.3, c] for c in color_attr]

    # Make sure dirname exists
    if outfile is not None:
        mkdir(os.path.dirname(outfile))

    # Ovito stuff. Can be customized by client code.
    from ovito.pipeline import StaticSource, Pipeline
    from ovito.data import DataCollection, SimulationCell, Particles
    from ovito.modifiers import CreateBondsModifier

    data = DataCollection()
    _cell = SimulationCell(pbc=(False, False, False))
    _cell[0, 0] = cell.side[0]
    _cell[1, 1] = cell.side[1]
    _cell[2, 2] = cell.side[2]
    _cell[:, 3] = -cell.side/2
    _cell.vis.enabled = True
    _cell.vis.line_width = 0.02
    # _cell.vis.rendering_color = (1.0, 1.0, 1.0)
    data.objects.append(_cell)

    # Create a Particles object
    # Scale radii by input radius
    particles = Particles()
    particles.create_property('Position', data=[p.position for p in particle])
    particles.create_property('Radius', data=[p.radius*2*radius for p in particle])
    particles.create_property('Color', data=colors)  # [[1,0,0], [1,0,0], [1,0,0]])
    data.objects.append(particles)

    # Create a new Pipeline with a StaticSource as data source:
    pipeline = Pipeline(source=StaticSource(data=data))

    # Apply client code callback
    if callback:
        callback(pipeline)
    pipeline.add_to_scene()

    # Define viewport
    if viewport:
        vp = viewport
    else:
        if perspective:
            vp = Viewport(type=Viewport.Type.Perspective, camera_dir=camera_dir, camera_pos=camera_pos)
        else:
            vp = Viewport(type=Viewport.Type.Ortho, camera_dir=camera_dir, camera_pos=camera_pos)

    # Render
    if zoom:
        vp.zoom_all()
    if outfile is None:
        fh = tempfile.NamedTemporaryFile('w', dir=tmpdir, suffix='.png', delete=False)
        outfile = fh.name

    vp.render_image(filename=outfile,
                    size=size,
                    renderer=TachyonRenderer())

    # Scene is a singleton, so we must clear it
    pipeline.remove_from_scene()

    # Try to display the image (e.g. in a jupyter notebook)
    try:
        from IPython.display import Image
        return Image(outfile)
    except ImportError:
        return outfile



# This file is part of atooms
# Copyright 2010-2024, Daniele Coslovich

"""LAMMPS trajectory format."""

import numpy

from .base import TrajectoryBase
from .folder import TrajectoryFolder
from atooms.system import System, Particle, Cell
from atooms.system.particle import distinct_species
from atooms.system.interaction import Interaction

import sys

# Redefine range for python
if sys.version_info[0] == 2:
    range = xrange  # noqa: F821 pylint: disable=undefined-variable

# Formatting callbacks

def _parse_type(data, idx, system):
    system.particle[idx].species = data

def _parse_x(data, idx, system):
    system.particle[idx].position[0] = float(data)

def _parse_y(data, idx, system):
    system.particle[idx].position[1] = float(data)

def _parse_z(data, idx, system):
    system.particle[idx].position[2] = float(data)

def _parse_xs(data, idx, system):
    system.particle[idx].position[0] = (float(data) - 0.5) * system.cell.side[0]

def _parse_ys(data, idx, system):
    system.particle[idx].position[1] = (float(data) - 0.5) * system.cell.side[1]

def _parse_zs(data, idx, system):
    system.particle[idx].position[2] = (float(data) - 0.5) * system.cell.side[2]

def _parse_vx(data, idx, system):
    system.particle[idx].velocity[0] = float(data)

def _parse_vy(data, idx, system):
    system.particle[idx].velocity[1] = float(data)

def _parse_vz(data, idx, system):
    system.particle[idx].velocity[2] = float(data)

def _parse_fx(data, idx, system):
    system.interaction.forces[idx, 0] = float(data)

def _parse_fy(data, idx, system):
    system.interaction.forces[idx, 1] = float(data)

def _parse_fz(data, idx, system):
    system.interaction.forces[idx, 2] = float(data)

def _parse_ix(data, idx, system):
    system.particle[idx]._ix = int(data)

def _parse_iy(data, idx, system):
    system.particle[idx]._iy = int(data)

def _parse_iz(data, idx, system):
    system.particle[idx]._iz = int(data)

def _parse_mass(data, idx, system):
    system.particle[idx].mass = float(data)

def _parse_charge(data, idx, system):
    system.particle[idx].charge = float(data)

def _parse_energy(data, idx, system):
    system.particle[idx].energy = float(data)

def _parse_molecule_id(data, idx, system):
    system.particle[idx].molecule_id = int(data)


class TrajectoryLAMMPS(TrajectoryBase):

    """
    LAMMPS format (https://docs.lammps.org/dump.html)

    In write mode, an additional .inp file is used as startup file.
    """

    suffix = 'atom'
    _cbk = {'x': _parse_x, 'y': _parse_y, 'z': _parse_z,
            'xu': _parse_x, 'yu': _parse_y, 'zu': _parse_z,
            'xs': _parse_xs, 'ys': _parse_ys, 'zs': _parse_zs,
            'xsu': _parse_xs, 'ysu': _parse_ys, 'zsu': _parse_zs,
            'vx': _parse_vx, 'vy': _parse_vy, 'vz': _parse_vz,
            'fx': _parse_fx, 'fy': _parse_fy, 'fz': _parse_fz,
            'ix': _parse_ix, 'iy': _parse_iy, 'iz': _parse_iz,
            'c_pe': _parse_energy, 'q': _parse_charge,
            'type': _parse_type, 'mass': _parse_mass, 'mol': _parse_molecule_id}

    def __init__(self, filename, mode='r', single_frame=False,
                 first_particle=-1, last_particle=-1, ignore_id=False):
        super(TrajectoryLAMMPS, self).__init__(filename, mode)
        self.precision = 14  # default to double precision
        self.single_frame = single_frame
        self.first_particle = first_particle
        self.last_particle = last_particle
        self.ignore_id = ignore_id
        self._file = open(self.filename, self.mode)
        if mode == 'r':
            self._setup_index()

    def _setup_index(self):
        """Sample indexing via tell / seek"""
        from collections import defaultdict
        self._file.seek(0)
        self._index_db = defaultdict(list)
        while True:
            data = self._file.readline()
            # We break if file is over or we found an empty line
            if not data:
                break
            if data.startswith('ITEM:'):
                for block in ['TIMESTEP', 'NUMBER OF ATOMS',
                              'BOX BOUNDS', 'ATOMS']:
                    if data[6:].startswith(block):
                        # entry contains whatever is found after block
                        entry = data[7+len(block):]
                        line = self._file.tell() - len(data)
                        self._index_db[block].append((line, entry))
                        break
                # Avoid reading after ATOOMS block has been found.
                # We assume it is the last block in the file.
                # The single_frame variable is a hint that there are
                # no more frames in the file
                if block == 'ATOMS' and self.single_frame:
                    break
        self._file.seek(0)

    def read_steps(self):
        steps = []
        for idx, _ in self._index_db['TIMESTEP']:
            self._file.seek(idx)
            self._file.readline()
            step = int(self._file.readline())
            steps.append(step)
        self._file.seek(0)
        return steps

    def read_system(self, frame):
        # Read number of particles
        idx, _ = self._index_db['NUMBER OF ATOMS'][frame]
        self._file.seek(idx)
        self._file.readline()
        data = self._file.readline()
        npart = int(data)

        # Build the system
        system = System()
        # First fill in the right number of particles.
        # This is necessary to be able to read them when unsorted
        system.particle = []
        for i in range(npart):
            if self.first_particle > 0 and i < self.first_particle:
                continue
            if self.last_particle > 0 and i >= self.last_particle:
                break
            system.particle.append(Particle())

        # Add cell
        idx, data = self._index_db['BOX BOUNDS'][frame]
        self._file.seek(idx)
        self._file.readline()
        ndim = len(data.split())
        # TODO: parse periodicity
        if ndim == 3:
            # Restricted triclinic: data == 'xy xz yz'
            # Orthogonal cell with PBC: data == 'pp pp pp'
            pass
        if ndim == 6:
            # Restricted triclinic with periodicity info
            ndim = 3
        if ndim == 2:
            # Triclinic, data == 'abc origin'
            raise ValueError('triclinic cells are not supported')

        L, center = [], []
        for _ in range(ndim):
            data = [float(x) for x in self._file.readline().split()]
            if len(data) == 3:
                # Currently triclinic is only supported if not tilted
                if float(data[2]) > 0.0:
                    raise ValueError('tilt not supported')
            L.append(data[1] - data[0])
            center.append((data[1] + data[0]) / 2)
        system.cell = Cell(numpy.array(L), center=numpy.array(center))

        # Read atoms data
        idx, data = self._index_db['ATOMS'][frame]
        fields = data.split()  # fields on a line
        _ = self._file.readline()

        # Add interaction if forces are present
        # In atooms, forces belong to the interaction, not to particles
        if 'fx' in fields or 'fy' in fields or 'fz' in fields:
            # TODO: this won't work with first and last particles
            system.interaction = Interaction()
            system.interaction.forces = numpy.ndarray((npart, ndim))
        else:
            interaction = None

        self.variables = []
        for field in fields:
            if field in ['x', 'xu', 'xs', 'xsu']:
                self.variables.append('particle.position')
            if field == 'vx':
                self.variables.append('particle.velocity')
            if field == 'q':
                self.variables.append('particle.charge')
            if field == 'mass':
                self.variables.append('particle.mass')
            if field == 'type':
                self.variables.append('particle.species')
        self.variables = tuple(self.variables)

        for i in range(npart):
            # Limit reading the ATOMS section if requested
            if self.first_particle > 0 and i < self.first_particle:
                continue
            if self.last_particle > 0 and i >= self.last_particle:
                break
            data = self._file.readline().split()
            # Accept unsorted particles by parsing their id
            if 'id' in fields and not self.ignore_id:
                idx = int(data[0]) - 1
            else:
                idx = i
            # Populate particle's attributes by reading fields
            for j, field in enumerate(fields):
                if field in self._cbk:
                    self._cbk[field](data[j], idx, system)

        return system

    def write_init(self, system):
        assert system.number_of_dimensions in [2, 3], 'can only use lammps with d=2,3'
        f = open(self.filename + '.inp', 'w')
        np = len(system.particle)
        L = system.cell.side
        species_db = distinct_species(system.particle)

        # LAMMPS header
        h = '\n'
        h += "{:d} atoms\n".format(np)
        h += "{:d} atom types\n".format(len(species_db))
        h += "{:.{prec}f} {:.{prec}f} xlo xhi\n".format(-L[0]/2, L[0]/2, prec=self.precision)
        h += "{:.{prec}f} {:.{prec}f} ylo yhi\n".format(-L[1]/2, L[1]/2, prec=self.precision)
        if system.number_of_dimensions == 3:
            h += "{:.{prec}f} {:.{prec}f} zlo zhi\n".format(-L[2]/2, L[2]/2, prec=self.precision)

        # LAMMPS body
        # Masses of species
        m = "\nMasses\n\n"
        for isp in range(len(species_db)):
            # Iterate over particles. Find instances of species and get masses
            for p in system.particle:
                if p.species == species_db[isp]:
                    m += '{:d} {:{prec}f}\n'.format(isp + 1, p.mass, prec=self.precision)
                    break

        # Atom coordinates
        r = "\nAtoms\n\n"
        atom_style = 'atomic'
        # Setup string formatting.
        # This complicated thing is to handle both 2d/3d
        fmt = '{:d} {:d}'  # index and species
        fmt_vel = '{:d}'
        if hasattr(system.particle[0], 'charge'):
            atom_style = 'charge'
            fmt += ' {}'
        if system.number_of_dimensions == 3:
            fmt += ' {:{prec}} {:{prec}} {:{prec}}'
            fmt_vel += ' {:{prec}} {:{prec}} {:{prec}}'
        else:
            fmt += ' {:{prec}} {:{prec}} 0.0'
            fmt_vel += ' {:{prec}} {:{prec}} 0.0'

        # Add internal image info, if present
        fmt += ' {} {} {}'
        _has_image = False
        if hasattr(system.particle[0], '_ix'):
            _has_image = True

        def _image(p):
            if _has_image:
                return [p._ix, p._iy, p._iz]
            return [0, 0, 0]

        fmt += '\n'
        fmt_vel += '\n'

        if atom_style == 'atomic':
            for i, p in enumerate(system.particle):
                isp = species_db.index(p.species) + 1
                r += fmt.format(i+1, isp, *p.position, *_image(p), prec=self.precision)
        elif atom_style == 'charge':
            for i, p in enumerate(system.particle):
                isp = species_db.index(p.species) + 1
                r += fmt.format(i+1, isp, p.charge, *p.position, *_image(p), prec=self.precision)

        v = "\nVelocities\n\n"
        for i, p in enumerate(system.particle):
            v += fmt_vel.format(i+1, *p.velocity, prec=self.precision)

        f.write(h)
        f.write(m)
        f.write(r)
        f.write(v)
        f.close()

    def write_system(self, system, step):
        pass

    def close(self):
        self._file.close()


class TrajectoryFolderLAMMPS(TrajectoryFolder):

    """
    Multi-file layout LAMMPS format.

    It looks for all files matching a pattern in the input folder path.
    """

    def __init__(self, filename, mode='r', file_pattern='*',
                 step_pattern=r'[a-zA-Z\.]*(\d*)', first_particle=-1, last_particle=-1):
        super(TrajectoryFolderLAMMPS, self).__init__(filename, mode=mode,
                                                     file_pattern=file_pattern,
                                                     step_pattern=step_pattern)
        self.first_particle = first_particle
        self.last_particle = last_particle
        # Small trick to force reading steps from lammps file
        self._steps = None
        # Sort frames according to step read in lammps file
        sorted_steps = sorted(self.steps)
        files_with_steps = [(x, y) for x, y in zip(self.files, self.steps)]
        files_with_steps.sort(key=lambda x: sorted_steps.index(x[1]))
        files = [_[0] for _ in files_with_steps]
        self.files = files
        self._steps = sorted_steps

    def read_steps(self):
        steps = []
        for filename in self.files:
            with TrajectoryLAMMPS(filename, 'r', single_frame=True,
                                  first_particle=self.first_particle,
                                  last_particle=self.last_particle) as th:
                steps.append(th.steps[0])
        return steps

    def read_system(self, frame):
        with TrajectoryLAMMPS(self.files[frame], 'r', single_frame=True,
                              first_particle=self.first_particle,
                              last_particle=self.last_particle) as th:
            return th[0]

    def write_system(self, system, step):
        # We cannot write
        raise NotImplementedError('cannot write lammps folder trajectory')



#!/usr/bin/env python

import os
import unittest
from atooms.trajectory import TrajectoryXYZ, TrajectoryRam
from atooms.simulation import Simulation
from atooms.core.utils import rmd, rmf
import atooms.backends.lammps
from atooms.backends.lammps import LAMMPS, Interaction
from atooms.system import System

atooms.backends.lammps.lammps_command = 'lmp'
SKIP = not atooms.backends.lammps.installed()

class Test(unittest.TestCase):

    def setUp(self):
        if SKIP:
            self.skipTest('missing LAMMPS')
        self.input_file = os.path.join(os.path.dirname(__file__),
                                       '../data/lj_N1000_rho1.0.xyz')

    def test_single(self):
        cmd = """
        pair_style      lj/cut 2.5
        pair_coeff      1 1 1.0 1.0 2.5
        neighbor        0.3 bin
        neigh_modify    every 20 delay 0 check no
        fix             1 all nve
        """
        bck = LAMMPS(self.input_file, cmd)
        sim = Simulation(bck)
        x = sim.system.particle[0].position[0]
        self.assertAlmostEqual(x, 3.62635, places=5)
        sim.run(10)
        x = sim.system.particle[0].position[0]
        self.assertAlmostEqual(x, 3.64526, places=5)
        sim.run(10)
        x = sim.system.particle[0].position[0]
        self.assertAlmostEqual(x, 3.675987, places=5)

    def test_images(self):
        cmd = """
        pair_style      lj/cut 2.5
        pair_coeff      1 1 1.0 1.0 2.5
        neighbor        0.3 bin
        neigh_modify    every 20 delay 0 check no
        fix             1 all nve
        """
        s = System(N=100)
        s.density = 0.3
        s.temperature = 2.0
        bck = LAMMPS(s, cmd)
        sim = Simulation(bck)
        sim.run(1000)
        ixs = sim.system.dump('_ix')
        sim.run(1)
        self.assertEqual(list(sim.system.dump('_ix')), list(ixs))
        # print(sim.system.dump('position')[:30, 0]-sim.system.dump('position_unfolded')[:30, 0])

    def test_nvt(self):
        import random
        cmd = """
        pair_style      lj/cut 2.5
        pair_coeff      1 1 1.0 1.0 2.5
        neighbor        0.3 bin
        neigh_modify    every 20 delay 0 check no
        fix             1 all nvt temp 2.0 2.0 0.2
        timestep        0.002
        """

        def store(sim, T, U):
            T.append(sim.system.temperature)
            U.append(sim.system.potential_energy(per_particle=True))

        with TrajectoryXYZ(self.input_file) as th:
            system = th[-1]

        for inp in [self.input_file, TrajectoryXYZ(self.input_file),
                    system]:
            T, U = [], []
            random.seed(1)
            bck = LAMMPS(inp, cmd)
            sim = Simulation(bck)
            sim.system.temperature = 1.5
            sim.add(store, 500, T, U)
            sim.run(2000)
            ave = sum(T[3:]) / len(T[3:])
            self.assertAlmostEqual(ave, 2.0, places=1)
            if isinstance(inp, TrajectoryXYZ):
                inp.close()

    def test_nvt_2d(self):
        import random
        from atooms.system import System, Thermostat
        cmd = """
        pair_style      lj/cut 2.5
        pair_coeff      1 1 1.0 1.0 2.5
        neighbor        0.5 bin
        timestep        0.004
        """

        random.seed(1)
        system = System(N=500, d=2)
        system.temperature = 2.0
        system.density = 0.7
        system.thermostat = Thermostat(2.0)
        T, U = [], []

        def store(sim, T, U):
            T.append(sim.system.temperature)
            U.append(sim.system.potential_energy(per_particle=True))

        bck = LAMMPS(system, cmd)
        sim = Simulation(bck)
        sim.add(store, 500, T, U)
        sim.run(5000)
        ave = sum(T[3:]) / len(T[3:])
        self.assertAlmostEqual(ave, 2.0, places=1)

    def test_unfold(self):
        import numpy
        import random
        from atooms.trajectory import Unfolded
        from atooms.simulation.observers import write, write_config
        from atooms.system import System, Thermostat
        cmd = """
        pair_style      lj/cut 2.5
        pair_coeff      1 1 1.0 1.0 2.5
        neighbor        0.5 bin
        timestep        0.004
        """
        random.seed(1)
        system = System(N=100, d=2)
        system.temperature = 2.0
        system.density = 0.7
        system.thermostat = Thermostat(2.0)
        bck = LAMMPS(system, cmd)
        sim = Simulation(bck)
        th = TrajectoryRam()
        sim.add(write_config, 1000, trajectory=th)
        sim.run(10000)
        pos_unf = th[-1].dump('position_unfolded')
        th_unf = Unfolded(th)
        pos_unf_cls = th_unf[-1].dump('position')
        self.assertFalse(numpy.any(numpy.abs(pos_unf - pos_unf_cls) > 1e-10))
        self.assertAlmostEqual(sim.rmsd, 6.1, places=1)
        self.assertAlmostEqual(bck.partial_rmsd[0], 6.1, places=1)

    def test_nvt_nofix(self):
        import random
        from atooms.system import Thermostat
        cmd = """
        pair_style      lj/cut 2.5
        pair_coeff      1 1 1.0 1.0 2.5
        neighbor        0.3 bin
        neigh_modify    every 20 delay 0 check no
        timestep        0.002
        """
        random.seed(1)
        T = []

        def store(sim, T):
            T.append(sim.system.temperature)
        bck = LAMMPS(self.input_file, cmd)
        sim = Simulation(bck)
        sim.system.temperature = 1.4
        sim.system.thermostat = Thermostat(2.0)
        sim.add(store, 500, T)
        sim.run(4000)
        ave = sum(T[3:]) / len(T[3:])
        self.assertAlmostEqual(ave, 2.0, places=1)

    def test_energy(self):
        cmd = """
        pair_style      lj/cut 2.5
        pair_coeff      1 1 1.0 1.0 2.5
        """
        bck = LAMMPS(self.input_file, cmd)
        bck.system.interaction.compute("energy", bck.system.particle, bck.system.cell)
        self.assertEqual(bck.system.interaction.energy / len(bck.system.particle), -4.2446836)  # crosschecked

        # Relaxed FCC
        bck = LAMMPS(os.path.join(os.path.dirname(__file__),
                                  '../data/lj_fcc_N108.xyz'), cmd)
        bck.system.interaction.compute("forces", bck.system.particle, bck.system.cell)
        # Test norm of force per particle
        U = bck.system.potential_energy(per_particle=True)
        W = bck.system.force_norm(per_particle=True)
        P = bck.system.pressure
        self.assertAlmostEqual(U, -7.7615881, places=7)
        self.assertAlmostEqual(W, 4.2e-11, places=2)
        self.assertAlmostEqual(P, -3.3935748, places=7)

    def test_roundoff(self):
        cmd = """
        pair_style      lj/cut 2.5
        pair_coeff      1 1 1.0 1.0 2.5
        neighbor        0.3 bin
        neigh_modify    every 20 delay 0 check no
        fix             1 all nve
        """
        import numpy
        numpy.set_printoptions(precision=15)
        bck = LAMMPS(self.input_file, cmd)
        bck.run(10)
        x1 = bck.system.particle[3].position[0]
        bck.run(0)
        x2 = bck.system.particle[3].position[0]
        # The difference should be of order of machine precision
        # Unfortunately, without packing/unpacking data in binary
        # format, we cannot maintain coehrence at machine precision
        self.assertAlmostEqual(abs(x2-x1), 0.0, places=12)

    def tearDown(self):
        rmd('/tmp/test_lammps.d')
        rmf('/tmp/test_lammps*')


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

import unittest
import sys
import os
import numpy

from atooms.simulation import Simulation, target
try:
    import rumd
    from rumdSimulation import rumdSimulation
    SKIP = False
except ImportError:
    SKIP = True

if not SKIP:
    from atooms.backends.rumd import System, Trajectory
    from atooms.backends.rumd import RUMD as Backend

from atooms.core.utils import setup_logging


setup_logging(level=40)

xyz = """\
     3
ioformat=1 dt=0.005000000 boxLengths=6.34960421,6.34960421,6.34960421 numTypes=1 Nose-Hoover-Ps=-0.027281716 Barostat-Pv=0.000000000 mass=1.0000000 columns=type,x,y,z,vx,vy,vz
0       -3.159757      3.145206     -3.145651 1.0 0.0 -1.0
0       -2.986284      3.045374     -2.362755 0.0 1.0 0.0
0       -2.813011      2.380848     -1.037014 -1.0 -1.0 1.0
"""

xyz_2 = """\
     4
ioformat=1 dt=0.005000000 boxLengths=6.34960421,6.34960421,6.34960421 numTypes=2 Nose-Hoover-Ps=-0.027281716 Barostat-Pv=0.000000000 mass=1.0000000,2.000000000 columns=type,x,y,z,vx,vy,vz
0       -3.159757      3.145206     -3.145651 1.0 0.0 1.0
0       -2.986284      3.045374     -2.362755 0.0 1.0 1.0
0       -2.813011      2.380848     -1.037014 0.0 1.0 1.0
1       -1.813011      1.380848     -0.037014 1.0 1.0 0.0
"""

xyz_io2 = """\
4
ioformat=2 timeStepIndex=1234 numTypes=4 integrator=IntegratorNVT,0.00400000019,0.346464646,0.200000003,0 sim_box=RectangularSimulationBox,5.65945244,5.65945244,5.65945244 mass=1.0,0.57,1e20,1e20 columns=type,x,y,z,imx,imy,imz,vx,vy,vz,fx,fy,fz,pe,vir
0 1.859081030 0.200642005 -0.850564003 0 0 0 -0.956519008 -0.441635996 -0.600809991 0.0 0.0 0.0 0.0 0.0
0 1.673092008 0.005931000 0.235958993 0 0 0 -0.661297977 -0.039301001 0.306021005 0.0 0.0 0.0 0.0 0.0
0 -0.317158997 2.685316086 2.080347061 0 0 0 -0.229003996 -0.226451993 0.536197007 0.0 0.0 0.0 0.0 0.0
0 -0.468039989 -0.332848012 1.445088983 0 0 0 -0.344832987 -0.324809015 -0.497487992 0.0 0.0 0.0 0.0 0.0
"""

# TODO: introduce generic tests for backends


class TestBackendRUMD(unittest.TestCase):

    def setUp(self):
        if SKIP:
            self.skipTest('no rumd')

        self.dout = '/tmp/test_backends_rumd_out'
        self.finp = '/tmp/test_backends_rumd_in.xyz'
        with open(self.finp, 'w') as fh:
            fh.write(xyz)

        self.backend = Backend(self.finp)
        self.sim = self.backend.rumd_simulation
        self.backend.rumd_simulation.sample.SetOutputDirectory(self.dout)
        self.backend.rumd_simulation.suppressAllOutput = True
        p = rumd.Pot_LJ_12_6(cutoff_method=rumd.ShiftedPotential)
        p.SetVerbose(False)
        p.SetParams(0, 0, 1., 1., 2.5)
        self.backend.rumd_simulation.SetPotential(p)
        itg = rumd.IntegratorNVT(targetTemperature=2.0, timeStep=0.002)
        self.backend.rumd_simulation.SetIntegrator(itg)

        self.finp2 = '/tmp/test_backends_rumd_in2.xyz'
        with open(self.finp2, 'w') as fh:
            fh.write(xyz_2)
        self.sim2 = rumdSimulation(self.finp2, verbose=False)

        self.finp_io2 = '/tmp/test_backends_rumd_io2.xyz'
        with open(self.finp_io2, 'w') as fh:
            fh.write(xyz_io2)

        from atooms.core.utils import mkdir
        mkdir('/tmp/test_backends')
        self.finp_io2_base = '/tmp/test_backends/0000001.xyz'
        with open(self.finp_io2_base, 'w') as fh:
            fh.write(xyz_io2)

    def test_system(self):
        system = System(self.sim.sample)
        U = system.potential_energy()
        T = system.temperature
        Uref = 36.9236726612
        Tref = 6.0 / (9 - 3)
        # Note places is the number of decimal places, not significant digits, 4 is enough
        self.assertAlmostEqual(U, Uref, 4)
        self.assertAlmostEqual(T, Tref)

    def test_temperature_mass(self):
        system = System(self.sim2.sample)
        T = system.temperature
        Tref = 10.0 / (12 - 3)  # if we don't have the right masses this will fail
        self.assertAlmostEqual(T, Tref)

    def test_particle(self):
        system = System(self.sim.sample)
        p = system.particle
        ref = numpy.array([-3.1597569, 3.14520597, -3.1456511])
        self.assertLess(max(abs(p[0].position - ref)), 1e-6)

    def test_particle_mass(self):
        system = System(self.sim2.sample)
        p = system.particle
        for mref, m in zip(numpy.array([1., 1., 1., 2.]), [pi.mass for pi in p]):
            self.assertAlmostEqual(m, mref)

    def test_trajectory_one_step(self):
        t = Trajectory(self.dout + '_one_step', 'w')
        system = System(self.sim.sample)
        t.write(system, 0)
        t.close()

    def test_trajectory_read(self):
        from atooms.trajectory import TrajectoryRUMD
        with TrajectoryRUMD(self.finp_io2, 'r') as t:
            self.assertEqual(t.steps, [1234])

    def test_trajectory_read_base(self):
        from atooms.trajectory import TrajectoryRUMD
        with TrajectoryRUMD(self.finp_io2_base, 'r') as t:
            self.assertEqual(t.steps, [1])

    def test_simulation(self):
        s = Simulation(self.backend, self.dout, steps=1)
        system = System(self.sim.sample)
        s.system = system
        s.run()

    def test_rmsd(self):
        s = Simulation(self.backend, self.dout, steps=1)
        s.run()
        self.assertGreater(s.rmsd, 0.0)

    def test_target_rmsd(self):
        s = Simulation(self.backend, self.dout, steps=sys.maxsize)
        s.add(target, 10, 'rmsd', 0.3)
        s.run()
        self.assertGreater(s.current_step, 1)
        self.assertGreater(s.rmsd, 0.3)

    def test_checkpoint(self):
        s = Simulation(self.backend, self.dout, steps=10, checkpoint_interval=10)
        s.run()
        # TODO: this will fail, change test for existence of chk file
        # self.assertTrue(os.path.exists(s.trajectory.filename + '.chk'))

    def tearDown(self):
        from atooms.core.utils import rmf, rmd
        rmf('/tmp/test_backends*')
        rmd('/tmp/test_backends')


if __name__ == '__main__':
    unittest.main()



import os
import glob
import json

from .interaction import Interaction
from .neighbor_list import NeighborList
from .verlet_list import VerletList
from .trajectory import Trajectory

# TODO: these functions were moved from atooms-models, check whether they are useful

def _available():
    """Pretty print the available potentials"""
    print('Available potentials:')
    for potential in _database_potential:
        print('- ', potential)

    print('Available cutoffs:')
    for what in _database_cutoff:
        print('- ', what)

def _potential(name, parameters, args='-O3 -ffast-math', verbose=False):
    """Get a potential from database"""
    from .helpers import _normalize_path
    import f2py_jit

    path = _normalize_path(name)
    extra_args = '--opt="{} {}"'.format('-ffree-form -ffree-line-length-none ', args)
    uid = f2py_jit.build_module(path, extra_args=extra_args, verbose=verbose)
    module = f2py_jit.import_module(uid)
    module.potential.setup(**parameters)

    class Potential():
        pass
    potential = Potential()
    potential.parameters = parameters
    potential.compute = module.potential.compute
    # if vectorize:
    #     if hasattr(module.potential, 'compute_vector'):
    #         potential.compute = module.potential.compute_vector
    #     else:
    #         raise ValueError('cannot vectorize {}'.format(name))
    # else:
    #     potential.compute = module.potential.compute
    return potential

def _add(path):
    """
    Add all f90 files in `path` to the global `database`
    """
    for _path in glob.glob(os.path.join(path, '*.f90')):
        # Be forgiving
        if not os.path.exists(_path):
            continue
        # Potential name is file basename
        name = os.path.basename(_path)[:-4]
        # Determine whether this is a potential or a cutoff
        with open(_path) as fh:
            line = fh.readline()
            if 'potential' in line:
                _database_potential[name] = ''
            elif 'cutoff' in line:
                _database_cutoff[name] = ''


# Singleton
_database_potential = {}
_database_cutoff = {}

# By default, load all json files in module path
_add(os.path.join(os.path.dirname(__file__)))
